"""
spectral_energy_distribution.py
====================================
Module to represent a spectral energy distribution (SED).
"""

from .sampled_spectrum import SampledSpectrum


class SpectralEnergyDistribution(SampledSpectrum):
    """
    A spectral energy distribution.
    """

    def __init__(self, source_id, wl, flux):
        SampledSpectrum.__init__(self, source_id, wl)
        self.flux = flux



"""
photometry_data.py
====================================
Module to represent a photometry dataframe.
"""
from os.path import join
from pathlib import Path

from astropy.io import fits
from astropy.io.votable import from_table, writeto
from astropy.table import Table
from fastavro import parse_schema, writer
from fastavro.validation import validate_many

from .output_data import OutputData
from .utils import _add_ecsv_header, _build_photometry_header


class PhotometryData(OutputData):

    def __init__(self, data):
        super().__init__(data, None)

    def _save_avro(self, output_path, output_file):
        """
        Save the output photometry in AVRO format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """

        def build_field(keys):
            return [{'name': key, 'type': 'long'} if key == 'source_id' else {'name': key, 'type': 'float'} for key in
                    keys]

        phot_list = self.data.to_dict('records')
        schema = {
            'doc': 'Output photometry.',
            'name': 'Photometry',
            'namespace': 'photometry',
            'type': 'record',
            'fields': build_field(phot_list[0].keys()),
        }
        validate_many(phot_list, schema)
        parsed_schema = parse_schema(schema)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.avro')
        with open(output_path, 'wb') as output:
            writer(output, parsed_schema, phot_list)

    def _save_csv(self, output_path, output_file):
        """
        Save the output photometry in CSV format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        photometry_df = self.data
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.csv')
        photometry_df.to_csv(output_path, index=False)

    def _save_ecsv(self, output_path, output_file):
        """
        Save the output photometry in ECSV format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        photometry_df = self.data
        header_lines = _build_photometry_header(photometry_df.columns)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        photometry_df.to_csv(join(output_path, f'{output_file}.ecsv'), index=False)
        _add_ecsv_header(header_lines, output_path, output_file)

    def _save_fits(self, output_path, output_file):
        """
        Save the output photometry in FITS format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        photometry_df = self.data
        table = Table.from_pandas(photometry_df)
        hdu_list = list()
        hdr = fits.Header()
        primary_hdu = fits.PrimaryHDU(header=hdr)
        hdu_list.append(primary_hdu)
        hdu = fits.table_to_hdu(table)
        hdu_list.append(hdu)
        # Put all HDUs together
        hdul = fits.HDUList(hdu_list)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.fits')
        hdul.writeto(output_path, overwrite=True)

    def _save_xml(self, output_path, output_file):
        """
        Save the output photometry in XML/VOTABLE format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        photometry_df = self.data
        table = Table.from_pandas(photometry_df)
        votable = from_table(table)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.xml')
        writeto(votable, output_path)



"""
plot_spectra.py
====================================
Module to plot spectra.
"""
import warnings
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from numpy import ndarray

from .multi_absolute import MultiAbsolutePlotter
from .multi_xp import MultiXpPlotter
from .single import SinglePlotter


def __plot_multi(spectra_type: str, spectra: pd.DataFrame, sampling: ndarray, show_plot: bool,
                 output_path: Optional[Union[Path, str]], output_file: Optional[str], format: str, legend: bool,
                 save_file: bool):
    if 'AbsoluteSampledSpectrum' in spectra_type:
        plotter = MultiAbsolutePlotter(spectra, sampling, show_plot, output_path, output_file, format,
                                       legend, save_file)
    elif spectra_type == 'XpSampledSpectrum':
        plotter = MultiXpPlotter(spectra, sampling, show_plot, output_path, output_file, format, legend, save_file)
    else:
        raise ValueError('Unrecognised spectra type.')
    return plotter


def plot_spectra(spectra: pd.DataFrame, sampling: ndarray = None, multi: bool = False, show_plot: bool = True,
                 output_path: Optional[Union[Path, str]] = None, output_file: Optional[str] = None, format: str = None,
                 legend: bool = True):
    """
    Plot one or more spectra.

    Args:
        spectra (DataFrame): DataFrame of spectra to be plotted.
        sampling (ndarray): Sampling used to create the spectra.
        multi (bool): Generate a multiple subplots. Default value of False plots each spectrum in its own figure. If
            True, errors will not be plotted.
        show_plot (bool): Show plots if True.
        output_path (str): Path to the directory where the figures will be saved. E.g.: '/home/user/folder'
        output_file (str): Name of the output file without extension (e.g. 'my_plot').
        format (str): File format for the saved figure. Defaults to 'jpg'.
        legend (bool): Print legend. Valid only if multi is True.
    """
    if sampling is None:
        raise ValueError('A sampling is required.')
    # If the user passes an argument related to saving the file, then assume the user wants the plot to be saved
    save_file = bool(output_path) or bool(output_file) or bool(format)
    if not save_file and not show_plot:
        warnings.warn("No values have been provided to save the file: output_path, output_file or format. The "
                      "parameter show_plot is set to False. The function will not generate a plot.", stacklevel=2)
        return
    format = format if format else 'jpg'
    if multi:
        plotter = __plot_multi(spectra.attrs['data_type'].__name__, spectra, sampling, show_plot,
                               output_path, output_file, format, legend, save_file)
    else:
        plotter = SinglePlotter(spectra, sampling, show_plot, output_path, output_file, format, legend, save_file)
    plotter.plot_fig()



"""
parse_generic.py
====================================
Module to parse input files containing spectra.
"""
from os.path import splitext

import pandas as pd
from astropy.io.votable import parse_single_table
from astropy.table import Table

from gaiaxpy.core.generic_functions import array_to_symmetric_matrix, str_to_array
from .cast import _cast

valid_extensions = ['avro', 'csv', 'ecsv', 'fits', 'xml']


def _raise_key_error(column):
    raise KeyError(f'The columns in the input data do not match the expected ones. Missing column {column}.')


class InvalidExtensionError(ValueError):
    """
    Error raised when the extension of the input file is not valid. It inherits from ValueError.
    """

    def __init__(self):
        valid = ', '.join(valid_extensions)
        message = f'Valid extensions are: {valid}.'
        Exception.__init__(self, message)


class GenericParser(object):
    """
    Generic spectra parser.
    """

    def __init__(self):
        self.info_msg = 'Reading input file...'

    def get_parser(self, extension):
        """
        Choose the parser to use based on the extension.

        Args:
            extension (str): File extension including the dot (e.g.: '.csv').

        Returns:
            method: Parse method corresponding to the extension.

        Raises:
            InvalidExtensionError: If the extension is not valid.
        """
        if extension == 'avro':
            return self._parse_avro
        elif extension in ['csv', 'ecsv']:
            return self._parse_csv
        elif extension == 'fits':
            return self._parse_fits
        elif extension == 'xml':
            return self._parse_xml
        else:
            raise InvalidExtensionError()

    def parse_file(self, file_path, disable_info=False):
        """
        Parse the input file according to its extension.

        Args:
            file_path (str): Path to a file.
            disable_info (bool): Whether to disable the progress tracker or not.

        Returns:
            DataFrame: Pandas DataFrame representing the file.
            str: File extension ('.csv', '.fits', or '.xml').
        """
        if not disable_info:
            self.print_info_msg()
        extension = _get_file_extension(file_path)
        parser = self.get_parser(extension)
        parsed_data = _cast(parser(file_path))
        if not disable_info:
            self.print_info_msg(done=True)
        return parsed_data, extension

    def _parse_avro(self, avro_file):
        raise NotImplementedError('Method not implemented for base class.')

    def _parse_csv(self, csv_file, _array_columns=None, _matrix_columns=None, _usecols=None):
        """
        Parse the input CSV file and store the result in a pandas DataFrame.

        Args:
            csv_file (str): Path to a CSV file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list of tuples): List of tuples where the first element is the number of rows/columns of a
                square matrix which values are those contained in the second element of the tuple.
            _usecols (list): Columns to read.

        Returns:
            DataFrame: A pandas DataFrame representing the CSV file.
        """
        df = pd.read_csv(csv_file, comment='#', float_precision='round_trip', usecols=_usecols)
        if _array_columns:  # Pandas converters seemed slower
            for column in _array_columns:
                if column in df.columns:
                    df[column] = df[column].apply(lambda x: str_to_array(x))
        if _matrix_columns:
            for size_column, values_column in _matrix_columns:
                df[values_column] = df.apply(lambda row: array_to_symmetric_matrix(str_to_array(row[values_column]),
                                                                                   row[size_column]), axis=1)
        return df

    def _parse_fits(self, fits_file, _array_columns=None, _matrix_columns=None, _usecols=None):
        """
        Parse the input FITS file and store the result in a pandas DataFrame.

        Args:
            fits_file (str): Path to a FITS file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list of tuples): List of tuples where the first element is the number of rows/columns of a
                square matrix which values are those contained in the second element of the tuple.
            _usecols (list): Columns to read.

        Returns:
            DataFrame: A pandas DataFrame representing the FITS file.
        """
        table = Table.read(fits_file, format='fits')
        df = table.to_pandas()[_usecols] if _usecols else table.to_pandas()
        if _matrix_columns:
            for size_column, values_column in _matrix_columns:
                df[values_column] = df.apply(lambda row: array_to_symmetric_matrix(row[values_column],
                                                                                   row[size_column]), axis=1)
        return df

    def _parse_xml(self, xml_file, _array_columns=None, _matrix_columns=None, _usecols=None):
        """
        Parse the input XML file and store the result in a pandas DataFrame.

        Args:
            xml_file (str): Path to an XML file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list of tuples): List of tuples where the first element is the number of rows/columns of a
                square matrix which values are those contained in the second element of the tuple.
            _usecols (list): Columns to read.

        Returns:
            DataFrame: A pandas DataFrame representing the XML file.
        """
        table = parse_single_table(xml_file).to_table()
        # The columns argument of the parse_single_table function triggers an error in certain versions of Astropy,
        # so all columns are read first, and then the unused ones are removed.
        df = table.to_pandas()[_usecols] if _usecols else table.to_pandas()
        if _matrix_columns:
            for size_column, values_column in _matrix_columns:
                df[values_column] = df.apply(lambda row: array_to_symmetric_matrix(row[values_column],
                                                                                   row[size_column]), axis=1)
        return df

    def print_info_msg(self, done=False):
        msg = self.info_msg
        if done:
            msg = msg + ' Done!'
        print(msg, end='\r')


def _get_file_extension(file_path):
    """
    Get the extension of a file.

    Args:
        file_path (str): Path to a file.

    Returns:
        str: File extension (e.g.: '.csv')
    """
    _, file_extension = splitext(file_path)
    return file_extension[1:]



class ExtensionNotImplementedError(Exception):
    def __init__(self, extension):
        super().__init__(f"This functionality is not implemented for the file extension '{extension}'")


class InvalidBandError(Exception):
    def __init__(self, band):
        self.message = f"String {band} is not a valid band. Band must be either 'bp' or 'rp'."
        super().__init__(self.message)


class NoBandsAvailableError(Exception):
    def __init__(self):
        super().__init__('No bands present in the input data. At least one band must be available.')


class SelectorNotImplementedError(Exception):
    def __init__(self, reader='This'):
        super().__init__(f'{reader} reader does not implement the selector functionality.')



from os.path import splitext

from gaiaxpy.core.generic_functions import standardise_extension, cast_output
from gaiaxpy.core.input_validator import check_column_overwrite
from gaiaxpy.file_parser.parse_external import ExternalParser
from gaiaxpy.file_parser.parse_internal_continuous import InternalContinuousParser
from gaiaxpy.input_reader.required_columns import MANDATORY_INPUT_COLS, COV_INPUT_COLUMNS, CORR_INPUT_COLUMNS, \
    TRUNCATION_COLS


def external(requested_columns=None, additional_columns=None, selector=None, **kwargs):
    return ExternalParser(requested_columns=requested_columns, additional_columns=additional_columns,
                          selector=selector, **kwargs)


def internal_continuous(requested_columns=None, additional_columns=None, selector=None, **kwargs):
    return InternalContinuousParser(requested_columns=requested_columns, additional_columns=additional_columns,
                                    selector=selector, **kwargs)


def raise_error():
    raise ValueError('File parser not implemented. This function cannot receive a file as input.')


covariance_extensions = ['avro']

function_parser_dict = {'apply_colour_equation': raise_error,
                        'convert': internal_continuous,
                        '_calibrate': internal_continuous,
                        'calibrate': internal_continuous,
                        '_generate': internal_continuous,
                        'generate': internal_continuous,
                        'get_inverse_covariance_matrix': internal_continuous,
                        'get_inverse_square_root_covariance_matrix': internal_continuous}


class FileReader:

    def __init__(self, file_parser_selector, file, truncation, additional_columns=None, selector=None,
                 disable_info=False, **kwargs):
        self.fps = file_parser_selector
        self.file = file
        self.file_extension = standardise_extension(splitext(file)[1])
        self.truncation = truncation
        self.additional_columns = dict() if additional_columns is None else additional_columns
        self.selector = selector
        self.disable_info = disable_info
        mandatory_columns = MANDATORY_INPUT_COLS.get(self.fps.function_name, list())
        style_columns = list()
        if mandatory_columns:
            # Files can contain covariances or correlations depending on the extension
            style_columns = COV_INPUT_COLUMNS if self.file_extension in covariance_extensions else CORR_INPUT_COLUMNS
        self.required_columns = mandatory_columns + style_columns
        if truncation:
            self.required_columns = self.required_columns + TRUNCATION_COLS
        self.requested_columns = self.required_columns
        if self.additional_columns:
            check_column_overwrite(additional_columns, self.required_columns)
            self.requested_columns = self.required_columns + self.get_extra_columns_from_extension()
        if kwargs:
            self.address = kwargs.get('address', None)
            self.port = kwargs.get('port', None)

    def get_extra_columns_from_extension(self):
        if self.file_extension == 'avro':
            return [c for c in self.additional_columns.keys() if c not in self.required_columns]
        else:
            # Verify columns
            for value in self.additional_columns.values():
                if not isinstance(value, list) or len(value) != 1:
                    if not value[0]:
                        raise ValueError('Empty values are not allowed in additional columns argument.')
                    else:
                        raise ValueError('Column lists longer than one are not allowed in non-nested formats.')
            return [self.additional_columns[c][0] for c in self.additional_columns.keys() if c not in
                    self.required_columns]

    def read(self):
        parser_arguments = {
            'requested_columns': self.requested_columns,
            'additional_columns': self.additional_columns,
            'selector': self.selector
        }
        if hasattr(self, 'address') and hasattr(self, 'port'):
            parser_arguments['address'] = self.address
            parser_arguments['port'] = self.port
        data, extension = self.fps.parser(**parser_arguments).parse_file(self.file, disable_info=self.disable_info)
        return cast_output(data), extension


class FileParserSelector(object):

    def __init__(self, function):
        self.function_name = function.__name__
        self.mandatory_columns = MANDATORY_INPUT_COLS.get(self.function_name, list())
        self.parser = function_parser_dict[self.function_name]



pbar_colour = '#509faf'
pbar_message = {'calibrator': 'Calibrating data',
                'colour_eq': 'Applying colour equation',
                'converter': 'Converting data',
                'correction': 'Correcting systems',
                'photometry': 'Generating photometry'}
pbar_units = {'calibrator': 'spec',
              'colour_eq': 'syst',
              'converter': 'spec',
              'correction': 'syst',
              'photometry': 'spec'}



"""
photometric_system.py
====================================
Module for the management of photometric systems.
"""

from configparser import ConfigParser
from os import remove
from os.path import exists

from aenum import Enum

from gaiaxpy.core.generic_functions import _get_built_in_systems, _is_built_in_system
from .config import _CFG_FILE_PATH, create_config, get_additional_filters_names, contains_filter_key
from .regular_photometric_system import RegularPhotometricSystem
from .standardised_photometric_system import StandardisedPhotometricSystem
from .utils import get_yes_no_answer


def get_available_systems():
    systems_list = _get_available_systems()
    return ', '.join(systems_list)


class AutoName(Enum):

    def get_system_name(self):
        return self.name

    def get_system_label(self):
        return self.value.label

    def get_zero_points(self):
        return self.value.zero_points

    def get_bands(self):
        return self.value.bands

    def get_offsets(self):
        return self.value.offsets

    def get_version(self):
        return self.value.version


def _system_is_standard(system_name):
    """
    Tell whether the input system is standard or not.

    Args:
        system_name (str): Photometric system name.

    Returns:
        bool: True is system is standard, false otherwise.
    """
    std_substring = system_name.split('_')[-1] if _is_built_in_system(system_name) else system_name[-3:]
    return std_substring.lower() == 'std'


def create_system(name, systems_path=None):
    return StandardisedPhotometricSystem(name, systems_path) if _system_is_standard(name) \
        else RegularPhotometricSystem(name, systems_path)


def _get_available_systems(config_file=None):
    """
    Get the available photometric systems according to the
    package configuration.

    Returns:
        str: A string containing the names of the photometric
             systems separated by spaces.
    """
    built_in_systems = _get_built_in_systems()
    # Try to load the configuration and see whether more systems have been defined
    additional_systems = get_additional_filters_names(config_file)
    return built_in_systems + additional_systems


def _get_system_tuples():
    return [(s, create_system(s, None)) if _is_built_in_system(s) else (s, create_system(s, _CFG_FILE_PATH)) for s in
            _get_available_systems(_CFG_FILE_PATH)]


system_tuples = _get_system_tuples()
PhotometricSystem = AutoName('PhotometricSystem', system_tuples)
PhotometricSystem.get_available_systems = get_available_systems


def get_current_filters_path():
    _config_parser = ConfigParser()
    _config_parser.read(_CFG_FILE_PATH)
    return _config_parser['filter']['filters_dir']


def load_additional_systems(_systems_path=None):
    """
    Load additional photometric systems.

    Args:
        _systems_path (str): Path to directory containing the additional filter files. If not provided, the program
        will ask the user to input one.

    Returns:
        Enum: PhotometricSystem object corresponding to an enumeration of the updated available systems.
    """
    updated_enum = __load_additional_systems(_systems_path, _CFG_FILE_PATH)
    updated_enum.get_available_systems = get_available_systems
    print('Systems loaded. Use PhotometricSystem.get_available_systems() to get the names of the current available'
          ' systems.')
    return updated_enum


def __load_additional_systems(_filters_path=None, config_file=None):
    """
    Load additional photometric systems. The names of these additional systems will start with the prefix 'USER.'
    For example, if a system is named 'X' in the loaded file, the corresponding photometric system will be referenced
    in GaiaXPy as 'USER_X', and it should be called using 'PhotometricSystem.USER_X,' where USER is not a username but
    simply the string 'USER'.

    Args:
        _filters_path (str): Path to directory containing the additional filter files.
        config_file (str): Path to configuration file where the path to the additional filter files will be stored.
    """
    config_file = config_file if config_file else _CFG_FILE_PATH
    if exists(config_file) and contains_filter_key(config_file):
        print(f'A path for additional filters has already been defined. The current path is '
              f'{get_current_filters_path()}')
        get_yes_no_answer('Do you want to redefine the path? [[y]/n]: ', yes_action=create_config, no_action=None,
                          yes_args=_filters_path)
    else:
        create_config(_filters_path, config_file)
    return AutoName('PhotometricSystem', _get_system_tuples())


def remove_additional_systems():
    """
    Remove previously loaded additional photometric systems. If no additional systems have been added, no changes will
    be made.

    Returns:
        Enum: PhotometricSystem object corresponding to an enumeration of the updated available systems.
    """
    if exists(_CFG_FILE_PATH):
        remove(_CFG_FILE_PATH)
        print('Additional systems configuration successfully removed.')
    else:
        print('No additional configuration exists.')
    _PhotometricSystem = AutoName('PhotometricSystem', _get_system_tuples())
    _PhotometricSystem.get_available_systems = get_available_systems
    return _PhotometricSystem



"""
config.py
====================================
Module to handle the calibrator and generator configuration files.
"""

from configparser import ConfigParser
from os.path import join

from gaiaxpy.config.paths import filters_path, config_ini_file
from gaiaxpy.core.satellite import BANDS
from gaiaxpy.core.xml_utils import get_file_root, get_array_text, get_xp_merge, get_xp_sampling_matrix

ADDITIONAL_SYSTEM_PREFIX = 'USER'


def get_file_path(config_file=None):
    if not config_file:
        return filters_path
    _config_parser = ConfigParser()
    _config_parser.read(config_file)
    try:
        file_path = _config_parser['filter']['filters_dir']
    except KeyError:
        return filters_path
    return file_path


def get_filter_version_from_config(_config_parser):
    try:
        version = _config_parser['filter']['version']
    except KeyError:
        version = None
    return version


def replace_file_name(_config_file, label, key, bp_model, rp_model, system):
    _config_parser = ConfigParser()
    _config_parser.read(_config_file)
    version = get_filter_version_from_config(_config_parser)
    if version:
        file_name = _config_parser.get(label, key).replace('version', version)
        system = system.replace(f'{ADDITIONAL_SYSTEM_PREFIX}_', '')
    else:
        file_name = _config_parser.get(label, key).format(label, key).replace('model',
                                                                              f'{bp_model}{rp_model}')
    file_name = file_name.replace('system', system) if system else file_name.replace('system_', '')
    return file_name


def get_file(label, key, system, bp_model, rp_model, config_file=None):
    """
    Get the file path corresponding to the given label and key.

    Args:
        label (str): Label of the photometric system or functionality (e.g.: 'Johnson' or 'calibrator').
        key (str): Type of file to load ('zeropoint', 'merge', 'sampling').
        system (str): Photometric system name, can be None in which case the generic configuration is loaded.
        bp_model (str): BP model.
        rp_model (str): RP model.
        config_file: Path to configuration file.

    Returns:
        str: Path of a file.
    """
    _config_file = config_ini_file if config_file is None else config_file
    file_name = replace_file_name(_config_file, label, key, bp_model, rp_model, system)
    file_path = get_file_path(config_file)
    return join(file_path, file_name)


def load_xpmerge_from_xml(system=None, bp_model=None, rp_model='v142r', config_file=None):
    """
    Load the XpMerge table from the filter XML file.

    Args:
        system (str): Name of the photometric system if it corresponds.
        bp_model (str): BP model.
        rp_model (str): RP model.
        config_file (str): Path to configuration file.

    Returns:
        ndarray: Array containing the sampling grid values.
        dict: A dictionary containing the XpMerge table with one entry for BP and one for RP.
    """
    bp_model = bp_model if bp_model else 'v375wi'
    label = key = 'filter'
    file_path = get_file(label, key, system, bp_model, rp_model, config_file=config_file)
    x_root = get_file_root(file_path)
    sampling_grid, bp_merge, rp_merge = get_xp_merge(x_root)
    return sampling_grid, dict(zip(BANDS, [bp_merge, rp_merge]))


def load_xpsampling_from_xml(system=None, bp_model=None, rp_model='v142r', config_file=None):
    """
    Load the XpSampling table from the XML filter file.

    Args:
        system (str): Photometric system name, can be None in which case the generic configuration is loaded.
        bp_model (str): BP model.
        rp_model (str): RP model.
        config_file (str): Path to configuration file.

    Returns:
        dict: A dictionary containing the XpSampling table with one entry for BP and one for RP.
    """
    bp_model = bp_model if bp_model else 'v375wi'
    xml_file = get_file('filter', 'filter', system, bp_model, rp_model, config_file=config_file)
    x_root = get_file_root(xml_file)
    _, n_bands = get_array_text(x_root, 'bands')

    bp_sampling = get_xp_sampling_matrix(x_root, 'bp', n_bands)
    rp_sampling = get_xp_sampling_matrix(x_root, 'rp', n_bands)

    xp_sampling = dict(zip(BANDS, [bp_sampling, rp_sampling]))
    return xp_sampling



from gaiaxpy.core.input_validator import check_column_overwrite
from gaiaxpy.input_reader.required_columns import CORR_INPUT_COLUMNS, MANDATORY_INPUT_COLS, TRUNCATION_COLS


class ArchiveReader(object):

    def __init__(self, function, truncation, user, password, additional_columns=None, disable_info=False):
        self.function = function
        self.truncation = truncation
        self.user = user
        self.password = password
        self.disable_info = disable_info
        self.info_msg = 'Running query...'
        # Columns
        self.additional_columns = dict() if additional_columns is None else additional_columns
        mandatory_columns = MANDATORY_INPUT_COLS.get(function.__name__, list())
        self.style_columns = CORR_INPUT_COLUMNS  # The Archive will always use correlations
        self.required_columns = list()
        if mandatory_columns:
            self.required_columns = mandatory_columns + self.style_columns
        if truncation:
            self.required_columns = self.required_columns + TRUNCATION_COLS
        self.requested_columns = self.required_columns
        if self.additional_columns:
            check_column_overwrite(additional_columns, self.required_columns)
            self.requested_columns = self.required_columns + [c for c in self.additional_columns.keys() if c not in
                                                              self.required_columns]

    def _login(self, gaia):
        user = self.user
        password = self.password
        if user and password:
            gaia.login(user=user, password=password)
        else:
            pass

    def show_info_msg(self, done=False):
        msg = self.info_msg
        if done:
            msg = msg + ' Done!'
        print(msg, end='\r')



from sys import stdout

from tqdm import tqdm

from gaiaxpy.core.generic_variables import pbar_colour, pbar_units, pbar_message
from gaiaxpy.spectrum.multi_synthetic_photometry import MultiSyntheticPhotometry
from .synthetic_photometry_generator import SyntheticPhotometryGenerator


class MultiSyntheticPhotometryGenerator(SyntheticPhotometryGenerator):

    def __init__(self, photometric_system, bp_model, rp_model):
        self.function_label = 'photsystem'
        if not photometric_system:
            raise ValueError('Photometric system list cannot be empty.')
        self.photometric_system = photometric_system
        self.system_label = [phot_system.get_system_label() for phot_system in self.photometric_system]
        self.bp_model = bp_model
        self.rp_model = rp_model

    def generate(self, parsed_input_data, extension, output_file, output_format, save_file, truncation):
        __FUNCTION_KEY = 'photometry'
        # Recover attributes
        systems = self.photometric_system
        internal_systems = [system.value for system in systems]
        # Generate XP variables
        xp_sampling_list = [system.load_xpsampling_from_xml() for system in internal_systems]
        xp_sampling_grid_xp_merge_tuples_list = [system.load_xpmerge_from_xml() for system in internal_systems]
        xp_sampling_grid_list = [element[0] for element in xp_sampling_grid_xp_merge_tuples_list]
        xp_merge_list = [element[1] for element in xp_sampling_grid_xp_merge_tuples_list]
        # Get basis functions list
        sampled_basis_func_list = [self._get_sampled_basis_functions(xp_sampling, xp_sampling_grid) for
                                   xp_sampling, xp_sampling_grid in zip(xp_sampling_list, xp_sampling_grid_list)]
        # One list per system
        photometry_list_of_lists = [self._create_photometry_list(parsed_input_data, phot_system,
                                                                 sampled_basis_func, truncation, xp_merge)
                                    for phot_system, sampled_basis_func, xp_merge
                                    in zip(systems, sampled_basis_func_list, xp_merge_list)]
        # Now the first list contains the photometries in all systems for the first source_id, and so on.
        rearranged_photometry_list = [sublist for sublist in tqdm(zip(*photometry_list_of_lists),
                                                                  desc=pbar_message[__FUNCTION_KEY],
                                                                  total=len(parsed_input_data),
                                                                  unit=pbar_units[__FUNCTION_KEY], leave=False,
                                                                  colour=pbar_colour, file=stdout)]
        return MultiSyntheticPhotometry(systems, rearranged_photometry_list)._generate_output_df()



"""
plotter.py
====================================
Module to create a plotter object.
"""

from os.path import join
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd

plt.rcParams['text.usetex'] = True


class Plotter:

    def __init__(self, spectra, sampling, show_plot, output_path, output_file, _format, legend, save_file):
        if isinstance(spectra, pd.Series):
            df = spectra.reset_index().transpose()
            df.columns = df.iloc[0]
            spectra = df[1:]
        self.spectra = spectra
        self.spectra_class = _set_class(spectra)
        self.sampling = sampling
        self.show_plot = show_plot
        self.output_path = output_path
        # Here we assume that the list only contains spectra of the same type
        self.output_file = self._set_output_file(output_file)
        self.format = _format
        self.legend = legend
        self.save_file = save_file
        # Set maximum number of spectra to plot
        self.max_spectra_on_multi = 40
        self.max_spectra_on_single = 20

    def plot_fig(self):
        raise NotImplementedError('Method not implemented for the parent class.')

    def _get_inputs(self, spectrum):
        return self.sampling, spectrum['flux'], spectrum['flux_error']

    def _get_source_id(self, spectrum):
        return spectrum['source_id']

    def _save_figure(self, index=None):
        """
        Args:
            index (int): Index to append at the end of the name of the figure, only used when many single plots are
            requested.
        """
        # Save the figure in the current path is no path is provided
        output_path = self.output_path if self.output_path else '.'
        # If index is provided, then append to file name
        output_file_name = f'{self.output_file}_{index}' if index is not None else self.output_file
        Path(output_path).mkdir(parents=True, exist_ok=True)
        plt.savefig(join(output_path, f'{output_file_name}.{self.format}'), format=self.format, transparent=False)

    def _set_output_file(self, output_file):
        return output_file if output_file else self.spectra_class.__name__


def _set_class(spectra):
    if isinstance(spectra, pd.DataFrame):
        return spectra.attrs['data_type']
    else:
        raise ValueError('Input should be a pandas DataFrame.')



"""
generic_functions.py
====================================
Module to hold some functions used by different subpackages.
"""

import sys
from ast import literal_eval
from collections import namedtuple
from os.path import join
from pathlib import Path
from string import capwords
from typing import Optional, Union, Tuple, List
from xml.etree import ElementTree as et

import numpy as np
import pandas as pd
from numpy import ndarray

from gaiaxpy.config.paths import filters_path, config_path
from gaiaxpy.core.custom_errors import InvalidBandError
from gaiaxpy.core.satellite import BANDS
from gaiaxpy.generator.config import get_additional_filters_path
from gaiaxpy.spectrum.utils import _correlation_to_covariance_dr3int5

# Verifying the function is valid first, will help when adding new functions
PHOTOMETRY_FUNCTIONS = ['generate', '_generate']
CHOLESKY_FUNCTIONS = ['get_inverse_covariance_matrix', 'get_inverse_square_root_covariance_matrix']
OTHER_FUNCTIONS = ['calibrate', 'convert']  # The ones that accept truncation and with_correlation arguments
ALL_ADD_COLS_FUNCTIONS = PHOTOMETRY_FUNCTIONS + CHOLESKY_FUNCTIONS + OTHER_FUNCTIONS


def _get_built_in_systems() -> list:
    av_sys = open(join(config_path, 'available_systems.txt'), 'r')
    return av_sys.read().splitlines()


def _is_built_in_system(system):
    return system in _get_built_in_systems()


def cast_output(output):
    cast_dict = {'source_id': 'int64', 'solution_id': 'int64'}
    cast_dict_keys = cast_dict.keys()
    df = output if isinstance(output, pd.DataFrame) else output.data
    for column in df.columns:
        if column in cast_dict_keys:
            df[column] = df[column].astype(cast_dict[column])
    return df


def parse_band(band):
    if isinstance(band, str):
        band = band.lower()
    elif isinstance(band, list) and len(band) == 1:
        band = band[0].lower()
    if band in BANDS:
        return band
    else:
        raise InvalidBandError(band)


def str_to_matrix(str_matrix):
    """
    Convert a string of the form ((1,2,3),(4,5,6),(7,8,9)) to a NumPy matrix.
    """
    # Replace nan to None so the string can be evaluated
    str_matrix = str_matrix.replace('nan', 'None')
    evaluated = literal_eval(str_matrix)
    evaluated = np.array(evaluated)
    evaluated = np.where(evaluated is None, np.nan, evaluated)
    return np.array(evaluated)


def str_to_array(str_array):
    if isinstance(str_array, np.ndarray):
        return str_array
    if isinstance(str_array, str) and len(str_array) >= 2 and str_array[0] == '(' and str_array[1] == '(':
        return str_to_matrix(str_array)
    elif isinstance(str_array, str):
        try:
            return np.fromstring(str_array[1:-1], sep=',')
        except Exception:  # np.fromstring may not raise an error but only show a warning depending on the version
            raise ValueError('Input cannot be converted to array.')
    elif isinstance(str_array, float):
        return float('NaN')
    else:
        raise ValueError('Unhandled type.')


def validate_pwl_sampling(sampling):
    # Receives a NumPy array. Validates sampling in pwl.
    min_sampling_value = -10
    max_sampling_value = 70
    if sampling is None:
        raise ValueError("Sampling can't be None.")
    if len(sampling) == 0:
        raise ValueError('Sampling must contain at least one point.')
    # Must be a numpy array
    if not isinstance(sampling, ndarray):
        raise TypeError('Sampling must be a NumPy array.')
    # Array must be sorted in ascending order
    if not np.array_equal(sampling, np.sort(sampling)):
        raise ValueError('Sampling must be in ascending order.')
    min_value = sampling[0]
    max_value = sampling[-1]
    if min_value < min_sampling_value or max_value > max_sampling_value:
        raise ValueError(f'Wrong value for sampling. Sampling accepts an array of values where the minimum value is '
                         f'{min_sampling_value} and the maximum is {max_sampling_value}.')


def validate_wl_sampling(sampling):
    min_value = 330
    max_value = 1050
    # Check sampling
    if sampling is not None:
        if sampling[0] >= sampling[-1]:
            raise ValueError('Sampling should be a non-decreasing array.')
        elif sampling[0] < min_value or sampling[-1] > max_value:
            raise ValueError(f'Wrong value for sampling. Sampling accepts an array of values where the minimum value '
                             f'is {min_value} and the maximum is {max_value}.')


def _warning(message):
    print(f'UserWarning: {message}', file=sys.stderr)


def get_spectra_type(spectra):
    """
    Get the spectra type.

    Args:
        spectra (object): A spectrum or a spectra list.

    Returns:
        str: Spectrum type (e.g. AbsoluteSampledSpectrum).
    """
    if isinstance(spectra, list):
        spectrum = spectra[0]
    elif isinstance(spectra, dict):
        spectrum = spectra[list(spectra.keys())[0]]
    else:
        spectrum = spectra
    return spectrum.__class__


def _get_system_label(name):
    """
    Get the label of the photometric system.

    Returns:
        str: A short description of the photometric system.
    """

    def snake_to_pascal(word):
        return capwords(word.replace("_", " ")).replace(" ", "")

    return snake_to_pascal(name) if _is_built_in_system(name) else name


def _get_system_path(is_built_in):
    return filters_path if is_built_in else get_additional_filters_path()


# AVRO files include the values in the diagonal, whereas others don't
def array_to_symmetric_matrix(array, array_size):
    """
    Convert the input 1D array into a 2D matrix. The array is assumed to store only the unique elements of a symmetric
        matrix (i.e. all elements above the diagonal plus the diagonal) in column major order. A full 2D matrix is
        returned symmetric with respect to the diagonal.

    Args:
        array (ndarray): 1D array.
        array_size (int): number of rows/columns in the output matrix.

    Returns:
        array of arrays: a full 2D matrix.

    Raises:
        TypeError: If array is not of type np.ndarray.
    """

    def contains_diagonal(_array_size, _array):
        return not len(_array) == len(np.tril_indices(_array_size - 1)[0])

    # Is NaN size or NaN array
    if pd.isna(array_size) or (isinstance(array, float) and pd.isna(array)):
        return array
    if isinstance(array_size, float):  # If the missing band source is present, floats may be returned when parsing
        float_size = float(array_size)
        array_size = int(array_size)
        if float_size != array_size:
            raise ValueError("Input must have a decimal part of exactly .0")
    if isinstance(array, np.ndarray):
        n_dim = array.ndim
        if array.size == 0 or n_dim == 2:  # Either empty or already a matrix
            return array
        elif n_dim == 1:
            array_size = int(array_size)
            matrix = np.zeros((array_size, array_size))
            np.fill_diagonal(matrix, 1.0)  # Add values in diagonal
            k = 0 if contains_diagonal(array_size, array) else -1  # Diagonal offset (from Numpy documentation)
            matrix[np.tril_indices(array_size, k=k)] = array
            transpose = matrix.transpose()
            transpose[np.tril_indices(array_size, -1)] = matrix[np.tril_indices(array_size, -1)]
            return transpose
    raise TypeError('Wrong argument types. Must be np.ndarray and integer or float.')


def _extract_systems_from_data(data_columns, photometric_system=None):
    if isinstance(photometric_system, list):
        return [system.get_system_label() for system in photometric_system]
    src = 'source_id'
    columns = list(data_columns.copy())
    if src in columns:
        columns.remove(src)
    if photometric_system:
        photometric_system = photometric_system if isinstance(photometric_system, list) else [photometric_system]
        systems = [system.get_system_label() for system in photometric_system]
    else:
        # Infer photometric_system from the data
        column_list = [column.split('_')[0] for column in columns]
        systems = list(dict.fromkeys(column_list))
    return systems


def correlation_from_covariance(covariance):
    if covariance is None:
        return None
    v = np.sqrt(np.diag(covariance))
    outer_v = np.outer(v, v)
    correlation = covariance / outer_v
    correlation[covariance == 0] = 0
    return correlation


def correlation_to_covariance(correlation: np.ndarray, error: np.ndarray, stdev: float) -> np.ndarray:
    """
    Compute the covariance matrix from the correlation values.

    If the input correlation values are a 1D array, the values are assumed to be the lower triangle of a
    symmetric matrix (excluding the diagonal). The array will be internally converted to a full correlation matrix.

    Args:
        correlation (ndarray): A 2D numpy array of shape (n, n) representing the correlation matrix.
            Alternatively, a 1D numpy array of length n*(n-1)/2 representing the lower triangle of a
            symmetric correlation matrix (excluding the diagonal).
        error (ndarray): A 1D numpy array of length n containing the flux errors.
        stdev (float): The scaling factor for the errors.

    Returns:
        ndarray: A 2D numpy array of shape (n, n) representing the covariance matrix.

    Raises:
        ValueError: If the dimensions of input correlation are not either 1 or 2.
    """
    if correlation is None or (isinstance(correlation, float) and pd.isna(correlation)):
        return None
    if correlation.ndim == 1:
        size = get_matrix_size_from_lower_triangle(correlation)
        new_matrix = np.zeros((size, size))
        new_matrix[np.tril_indices(size, k=-1)] = correlation
        new_matrix += new_matrix.T
        new_matrix += np.diag(np.ones(size), k=0)
        correlation = new_matrix
    if correlation.ndim == 2:
        return _correlation_to_covariance_dr3int5(correlation, error, stdev)
    else:
        raise ValueError('Dimensions of input correlation must be either 1 or 2.')


def get_matrix_size_from_lower_triangle(array):
    """
    Compute the size of a symmetric matrix from an array containing the values of its lower triangle, excluding the
    diagonal.

    Args:
        array (ndarray): An array of length n*(n-1)/2 representing the lower triangle of an n x n symmetric matrix,
            excluding the diagonal.

    Returns:
        int: The size n of the symmetric matrix.

    Raises:
        ValueError: If the length of array is not a valid input for a symmetric matrix lower triangle.
    """
    return int((np.sqrt(1 + 8 * len(array)) + 1) / 2)


def standardise_extension(_extension):
    """
    Standardise the provided extension which can contain or not an initial dot, and can contain a mix of uppercase and
    lowercase letters.

    Args:
        _extension (str): File extension which may or may not contain an initial dot.

    Returns:
        str: The extension in lowercase letters and with no initial dot (e.g.: 'csv').
    """
    try:
        _extension = _extension[1:] if _extension[0] == '.' else _extension  # Remove initial dot if present
        return _extension.lower()
    except IndexError:
        raise ValueError(f"Extension '{_extension}' could not be parsed appropriately.")


def reverse_simple_add_col_dict(d):
    def __reverse_error(v):
        raise ValueError(f'List length should be one, but is {len(v)}.')

    return {value[0]: key if len(value) == 1 else __reverse_error(value) for key, value in d.items()}


def validate_additional_columns(additional_columns, function, **kwargs):
    systems = None
    additional_columns = format_additional_columns(additional_columns)
    function_name = function.__name__
    if function_name not in ALL_ADD_COLS_FUNCTIONS:
        raise ValueError(f'Function {function_name} does not accept additional columns.')
    if function_name in PHOTOMETRY_FUNCTIONS:
        systems = kwargs['photometric_system']
    if systems and function.__name__ not in ['_generate', 'generate']:
        raise ValueError('Systems will only work with photometry-related functions.')


def convert_values_to_lists(d):
    for key, value in d.items():
        if not isinstance(value, list):
            d[key] = [value]
    return d


def format_additional_columns(additional_columns: Optional[Union[str, list, dict]]):
    """
    Ensure additional columns are in the expected format. Output should be a dictionary, values in the dictionary
    should become lists to work with nested AVRO keys.
    """
    if additional_columns is None:
        return dict()
    if isinstance(additional_columns, str):
        return {additional_columns: [additional_columns]}
    if isinstance(additional_columns, list):
        return {v: [v] for v in additional_columns}
    if isinstance(additional_columns, dict):
        return convert_values_to_lists(additional_columns)


def validate_photometric_system(photometric_system):
    """
    Ensure photometric system input isn't empty.
    """
    if photometric_system in (None, [], ''):
        raise ValueError('At least one photometric system is required as input.')


def rename_with_required(data, additional_columns):
    return data.rename(columns=reverse_simple_add_col_dict(additional_columns))


def is_array_empty(array):
    return array is None or (not isinstance(array, np.ndarray) and pd.isna(array)) or (isinstance(array, np.ndarray)
                                                                                       and len(array) == 0)


def is_variable_empty(var):
    return var is None or pd.isna(var)


def get_bands_config(bases_config):
    if hasattr(bases_config, 'hermiteFunction'):
        return bases_config.hermiteFunction
    elif hasattr(bases_config, 'basisDefinition'):
        return bases_config.basisDefinition.spline
    else:
        raise ValueError('Bases configuration type not implemented.')


def __range_to_output(d: dict) -> Tuple[float, float]:
    """
    Extract the 'from' and 'to' values from a dictionary `d` and return them as a tuple of two floats.

    :param d: A dictionary containing the keys 'from' and 'to', whose values are assumed to be convertible to floats.

    :return: A tuple of two floats representing the 'from' and 'to' values from the dictionary.

    :raises KeyError: If the 'from' or 'to' key is not present in the input dictionary.
    :raises ValueError: If the 'from' or 'to' value is not a valid float.
    """
    return float(d['from']), float(d['to'])


def __is_empty_string(string: str):
    def replace_empty_characters(s):
        separator_chars = ['\n', '\t']
        for ec in separator_chars:
            s = s.replace(ec, '')
        return s

    string = replace_empty_characters(string)
    for char in string:
        if not char.isspace():
            return False
    return True


def __parse_text_number(text):
    try:
        integer_value = int(text)
        return integer_value
    except ValueError:
        try:
            float_value = float(text)
            return float_value
        except ValueError:
            return text


def __add_element_to_dict(element: et.Element, output_dict: dict):
    """
    Add an element to a dictionary based on its tag and attributes.

    :param element: The element to add to the dictionary.
    :param output_dict: The dictionary to add the element to.

    :return: None

    :raises ValueError: If the element tag is not recognised.
    """
    if element.attrib:
        output_dict[element.tag] = __range_to_output(element.attrib)
    elif 'knots' in element.tag.lower():
        values = [float(value.text) for value in element.iter('value')]
        output_dict[element.tag] = tuple(values)
    # In some cases text is a string containing only a linebreak followed by blank spaces. In other cases, it can be
    # an integer or float as string, or plain text
    elif element.text:
        try:
            output_dict[element.tag] = float(element.text) if '.' in element.text else int(element.text)
        except (ValueError, TypeError):
            if __is_empty_string(element.text):
                output_dict[element.tag] = __parse_config(element, return_dict=True)
            else:
                output_dict[element.tag] = element.text
    else:
        raise ValueError(f'Unrecognized tag: {element.tag}')


def __iterative_find(x_root: et.Element, tag_list: List[str]) -> Optional[et.Element]:
    """
    Iteratively search for nested elements in an XML tree.
    :param x_root: The root element of the XML tree to search in.
    :type x_root: Element
    :param tag_list: A list of tag names to search for, in order.
    :type tag_list: List[str]
    :return: The first nested element that matches all tags in the `tag_list`, or `None` if no such element is found.
    :rtype: Element or None
    :raises: No exceptions are raised by this function.
    """

    def find_in_root(_x_root: et.Element, _tag: str) -> Optional[et.Element]:
        """
        Find the first child element of `_x_root` with the given `tag` name.
        :param _x_root: The root element to search in.
        :type _x_root: Element
        :param _tag: The tag name to search for.
        :type _tag: str
        :return: The first child element of `x_root` with the given `tag` name, or `None` if not found.
        :rtype: Element or None
        :raises: No exceptions are raised by this function.
        """
        return _x_root.find(_tag)

    result = x_root
    for tag in tag_list:
        result = find_in_root(result, tag)
        if result is None:
            break
    return result


def __generate_iterative_output(x_root, outer_tags):
    def is_matrix_or_knots(_tag):
        return 'matrix' in _tag.lower() or 'knots' == _tag.lower()

    _output = {}
    for tag in outer_tags:
        found_elements = __iterative_find(x_root, [tag])
        _output[tag] = dict()
        if len(found_elements) == 0 and not found_elements.attrib:
            _output[tag] = __parse_text_number(found_elements.text)
        elif len(found_elements) == 0 and found_elements.attrib:
            _output[tag] = __range_to_output(found_elements.attrib)
        elif is_matrix_or_knots(found_elements.tag):
            values = [float(value.text) for value in found_elements.iter('value')]
            _output[found_elements.tag] = np.array(values)
        else:
            for element in found_elements:
                __add_element_to_dict(element, _output[tag])
    return _output


def __generate_array_output(x_root, outer_tags):
    output_array = [el.text for el in x_root.findall(outer_tags[0])]
    try:
        output_array = tuple([float(el) for el in output_array])
    except ValueError:
        output_array = tuple(output_array)  # Elements can't be parsed to floats, may be actual strings
    return output_array


def __create_namedtuple(label: str, data: dict) -> namedtuple:
    """
    Create a named tuple with the given label and data.

    :param label: A string representing the label to use for the named tuple.
    :type label: str
    :param data: A dictionary representing the data to be stored in the named tuple.
    :type data: dict
    :return: A named tuple with the given label and data.
    :rtype: namedtuple
    """

    def __add_matrix_to_config(_data, _xp):
        config_str = f'{_xp}Config'
        hermite_function = _data.get('basisDefinition', {}).get('hermiteFunction', {})
        _config = hermite_function.get(config_str, {})
        if _config and 'transformationMatrix' not in _config:
            dimension = _config.get('dimension', None)
            if dimension:
                _config['transformationMatrix'] = np.identity(dimension)
                _config['transformedSetDimension'] = dimension

    for xp in ['bp', 'rp']:
        __add_matrix_to_config(data, xp)
    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = __create_namedtuple('Tuple', value)
    CalTuple = namedtuple(label, data)
    cal_tuple = CalTuple(**data)
    return cal_tuple


def __parse_config(x_root: et.Element, outer_tags: Optional[List[str]] = None, outer_title: Optional[str] = None,
                   return_dict=False) -> namedtuple:
    """
    Parse the configuration in x_root. If no outer_tags list is given, the function will parse every tag in the input
    x_root.
    """
    outer_tags = list(child.tag for child in x_root) if not outer_tags else outer_tags
    is_array = len(outer_tags) > 1 and len(set(outer_tags)) == 1
    _output = __generate_array_output(x_root, outer_tags) if is_array else __generate_iterative_output(x_root,
                                                                                                       outer_tags)
    outer_title = outer_title if outer_title else 'Internal'
    return _output if return_dict else __create_namedtuple(outer_title, _output)


def __get_file_root(xml_file: Union[Path, str]):
    """
    Parse an XML file and return its root element.

    :param xml_file: Path to the XML file to parse.
    :type xml_file: str

    :return: The root element of the parsed XML file.
    :rtype: Element

    :raises FileNotFoundError: If the specified XML file does not exist.
    :raises ElementTree.ParseError: If the specified XML file is not well-formed.
    """
    xtree = et.parse(xml_file)
    return xtree.getroot()


def parse_config(xml_file: Union[Path, str]) -> namedtuple:
    """
    Parses a configuration file in XML format and returns a named tuple with the settings. It returns the values for
    all the tags present in the file.

    :param xml_file: The path to the XML file to parse.
    :type xml_file: Union[Path, str]
    :return: A named tuple containing the configuration settings.
    :rtype: namedtuple
    """
    x_root = __get_file_root(xml_file)
    outer_title = x_root.tag.split('}')[1]
    return __parse_config(x_root, outer_title=outer_title)


def format_sampled_output(spectra_series, with_correlation):
    positions = spectra_series.iloc[0].get_positions()
    spectra_type = get_spectra_type(spectra_series.iloc[0])
    spectra_series = spectra_series.map(lambda x: x.spectrum_to_dict(with_correlation))
    spectra_df = pd.DataFrame(spectra_series.tolist())
    spectra_df.attrs['data_type'] = spectra_type
    return spectra_df, positions



"""
config.py
====================================
Module to work with the converter configuration.
"""
import xml.etree.ElementTree as et
from collections import namedtuple
from pathlib import Path
from typing import Tuple, Optional, List, Union

import numpy as np
import pandas as pd


def __convert_to_dict(obj):
    if obj.__class__.__name__ == 'Tuple':
        return {key: __convert_to_dict(val) for key, val in obj._asdict().items()}
    elif hasattr(obj, '__dict__'):
        return {key: __convert_to_dict(val) for key, val in obj.__dict__.items()}
    else:
        return obj


def __range_to_output(d: dict) -> Tuple[float, float]:
    """
    Extract the 'from' and 'to' values from a dictionary `d` and return them as a tuple of two floats.

    :param d: A dictionary containing the keys 'from' and 'to', whose values are assumed to be convertible to floats.

    :return: A tuple of two floats representing the 'from' and 'to' values from the dictionary.

    :raises KeyError: If the 'from' or 'to' key is not present in the input dictionary.
    :raises ValueError: If the 'from' or 'to' value is not a valid float.
    """
    return float(d['from']), float(d['to'])


def __is_empty_string(string: str):
    def replace_empty_characters(s):
        separator_chars = ['\n', '\t']
        for ec in separator_chars:
            s = s.replace(ec, '')
        return s

    string = replace_empty_characters(string)
    for char in string:
        if not char.isspace():
            return False
    return True


def __parse_text_number(text):
    try:
        integer_value = int(text)
        return integer_value
    except ValueError:
        try:
            float_value = float(text)
            return float_value
        except ValueError:
            return text


def __add_element_to_dict(element: et.Element, output_dict: dict):
    """
    Add an element to a dictionary based on its tag and attributes.

    :param element: The element to add to the dictionary.
    :param output_dict: The dictionary to add the element to.

    :return: None

    :raises ValueError: If the element tag is not recognised.
    """
    if element.attrib:
        output_dict[element.tag] = __range_to_output(element.attrib)
    elif 'knots' in element.tag.lower():
        values = [float(value.text) for value in element.iter('value')]
        output_dict[element.tag] = tuple(values)
    # In some cases text is a string containing only a linebreak followed by blank spaces. In other cases, it can be
    # an integer or float as string, or plain text
    elif element.text:
        try:
            output_dict[element.tag] = float(element.text) if '.' in element.text else int(element.text)
        except (ValueError, TypeError):
            if __is_empty_string(element.text):
                output_dict[element.tag] = __parse_config(element, return_dict=True)
            else:
                output_dict[element.tag] = element.text
    else:
        raise ValueError(f'Unrecognized tag: {element.tag}')


def __iterative_find(x_root: et.Element, tag_list: List[str]) -> Optional[et.Element]:
    """
    Iteratively search for nested elements in an XML tree.
    :param x_root: The root element of the XML tree to search in.
    :type x_root: Element
    :param tag_list: A list of tag names to search for, in order.
    :type tag_list: List[str]
    :return: The first nested element that matches all tags in the `tag_list`, or `None` if no such element is found.
    :rtype: Element or None
    :raises: No exceptions are raised by this function.
    """

    def find_in_root(_x_root: et.Element, _tag: str) -> Optional[et.Element]:
        """
        Find the first child element of `_x_root` with the given `tag` name.
        :param _x_root: The root element to search in.
        :type _x_root: Element
        :param _tag: The tag name to search for.
        :type _tag: str
        :return: The first child element of `x_root` with the given `tag` name, or `None` if not found.
        :rtype: Element or None
        :raises: No exceptions are raised by this function.
        """
        return _x_root.find(_tag)

    result = x_root
    for tag in tag_list:
        result = find_in_root(result, tag)
        if result is None:
            break
    return result


def __generate_iterative_output(x_root, outer_tags):
    def is_matrix_or_knots(_tag):
        return 'matrix' in _tag.lower() or 'knots' == _tag.lower()

    _output = {}
    for tag in outer_tags:
        found_elements = __iterative_find(x_root, [tag])
        _output[tag] = dict()
        if len(found_elements) == 0 and not found_elements.attrib:
            _output[tag] = __parse_text_number(found_elements.text)
        elif len(found_elements) == 0 and found_elements.attrib:
            _output[tag] = __range_to_output(found_elements.attrib)
        elif is_matrix_or_knots(found_elements.tag):
            values = [float(value.text) for value in found_elements.iter('value')]
            _output[found_elements.tag] = np.array(values)
        else:
            for element in found_elements:
                __add_element_to_dict(element, _output[tag])
    return _output


def __generate_array_output(x_root, outer_tags):
    output_array = [el.text for el in x_root.findall(outer_tags[0])]
    try:
        output_array = tuple([float(el) for el in output_array])
    except ValueError:
        output_array = tuple(output_array)  # Elements can't be parsed to floats, may be actual strings
    return output_array


def __create_namedtuple(label: str, data: dict) -> namedtuple:
    """
    Create a named tuple with the given label and data.

    :param label: A string representing the label to use for the named tuple.
    :type label: str
    :param data: A dictionary representing the data to be stored in the named tuple.
    :type data: dict
    :return: A named tuple with the given label and data.
    :rtype: namedtuple
    """

    def __add_matrix_to_config(_data, _xp):
        config_str = f'{_xp}Config'
        hermite_function = _data.get('basisDefinition', {}).get('hermiteFunction', {})
        _config = hermite_function.get(config_str, {})
        if _config and 'transformationMatrix' not in _config:
            dimension = _config.get('dimension', None)
            if dimension:
                _config['transformationMatrix'] = np.identity(dimension)
                _config['transformedSetDimension'] = dimension

    for xp in ['bp', 'rp']:
        __add_matrix_to_config(data, xp)
    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = __create_namedtuple('Tuple', value)
    CalTuple = namedtuple(label, data)
    cal_tuple = CalTuple(**data)
    return cal_tuple


def __parse_config(x_root: et.Element, outer_tags: Optional[List[str]] = None, outer_title: Optional[str] = None,
                   return_dict=False) -> namedtuple:
    """
    Parse the configuration in x_root. If no outer_tags list is given, the function will parse every tag in the input
    x_root.
    """
    outer_tags = list(child.tag for child in x_root) if not outer_tags else outer_tags
    is_array = len(outer_tags) > 1 and len(set(outer_tags)) == 1
    _output = __generate_array_output(x_root, outer_tags) if is_array else __generate_iterative_output(x_root,
                                                                                                       outer_tags)
    outer_title = outer_title if outer_title else 'Internal'
    return _output if return_dict else __create_namedtuple(outer_title, _output)


def __get_file_root(xml_file: Union[Path, str]):
    """
    Parse an XML file and return its root element.

    :param xml_file: Path to the XML file to parse.
    :type xml_file: str

    :return: The root element of the parsed XML file.
    :rtype: Element

    :raises FileNotFoundError: If the specified XML file does not exist.
    :raises ElementTree.ParseError: If the specified XML file is not well-formed.
    """
    xtree = et.parse(xml_file)
    return xtree.getroot()


def parse_config(xml_file: Union[Path, str]) -> namedtuple:
    """
    Parses a configuration file in XML format and returns a named tuple with the settings. It returns the values for
    all the tags present in the file.

    :param xml_file: The path to the XML file to parse.
    :type xml_file: Union[Path, str]
    :return: A named tuple containing the configuration settings.
    :rtype: namedtuple
    """
    x_root = __get_file_root(xml_file)
    outer_title = x_root.tag.split('}')[1]
    return __parse_config(x_root, outer_title=outer_title)


def get_bands_config(bases_config):
    if hasattr(bases_config, 'hermiteFunction'):
        return bases_config.hermiteFunction
    elif hasattr(bases_config, 'basisDefinition'):
        return bases_config.basisDefinition.spline
    else:
        raise ValueError('Bases configuration type not implemented.')


def get_unique_id(config: pd.DataFrame, _id: int) -> pd.DataFrame:
    """
    Access the group of rows in the configuration for the given ID.

    Args:
        config (DataFrame): A DataFrame containing the configuration columns and values.
        _id (int): An identifier of a set of configurations.

    Returns:
        DataFrame: A DataFrame containing the configuration values.
    """
    return config.loc[config['uniqueId'] == _id]



"""
utils.py
====================================
Module to hold methods useful for different kinds of spectra.
"""

import numpy as np
import pandas as pd


def get_covariance_matrix(row, band):
    columns = row.keys() if isinstance(row, dict) else row.index
    if f'{band}_covariance_matrix' in columns:
        covariance_matrix = row[f'{band}_covariance_matrix']
        return np.nan if covariance_matrix is None else covariance_matrix
    elif f'{band}_coefficient_covariances' in columns:
        coefficient_covariances = row[f'{band}_coefficient_covariances']
        return np.nan if coefficient_covariances is None else coefficient_covariances
    elif f'{band}_coefficient_correlations' in columns:
        return _correlation_to_covariance_dr3int5(row[f'{band}_coefficient_correlations'],
                                                  row[f'{band}_coefficient_errors'],
                                                  row[f'{band}_standard_deviation'])
    raise ValueError(f'None of the expected columns could be found in the input row. Columns are: {columns}.')


def _correlation_to_covariance_dr3int5(correlation_matrix, formal_errors, standard_deviation):
    """
    Compute the covariance matrix from the correlation matrix and the parameter formal errors.

    Args:
        correlation_matrix (ndarray): Correlation matrix, 2D array.
        formal_errors (ndarray): Formal errors of the parameters.
        standard_deviation (float): Standard deviation of the LSQ solution.

    Returns:
        ndarray: The covariance matrix as a numpy array.
    """
    if pd.isna(standard_deviation):
        return np.nan
    diagonal_errors = np.diag(formal_errors) / standard_deviation
    return diagonal_errors @ correlation_matrix @ diagonal_errors


def _correlation_to_covariance_dr3int4(correlation_matrix, formal_errors, standard_deviation):
    """
    Compute the covariance matrix from the correlation matrix and the parameter formal errors.

    Args:
        correlation_matrix (ndarray): Correlation matrix.
        formal_errors (ndarray): Formal errors of the parameters.
        standard_deviation (float): Standard deviation of the LSQ solution.

    Returns:
        ndarray: The covariance matrix as a numpy array.
    """
    diagonal_errors = np.diag(formal_errors) / standard_deviation
    correlation_matrix_aux = np.multiply(correlation_matrix, (standard_deviation * standard_deviation))
    np.fill_diagonal(correlation_matrix_aux, 1.)
    return diagonal_errors @ correlation_matrix_aux @ diagonal_errors


def _correlation_to_covariance_dr3int3(correlation_matrix, formal_errors):
    """
    Compute the covariance matrix from the correlation matrix and the parameter formal errors.

    Args:
        correlation_matrix (ndarray): Correlation matrix.
        formal_errors (ndarray): Formal errors of the parameters.

    Returns:
        ndarray: The covariance matrix as a numpy array.
    """
    # Populate the diagonal matrix containing the formal errors.
    diagonal_errors = np.diag(formal_errors)
    covariance_matrix = diagonal_errors @ correlation_matrix @ diagonal_errors
    return covariance_matrix


def _list_to_array(lst):
    """
    List to NumPy array.
    """
    if lst is None or (isinstance(lst, float) and pd.isna(lst)):
        return None
    if isinstance(lst, np.ndarray):
        return lst
    elif isinstance(lst, list):
        if not lst:
            raise ValueError('List cannot be empty.')
        else:
            return np.array(lst)
    raise ValueError('Wrong input type.')



"""
multi_xp.py
====================================
Module to plot multiple XP spectra.
"""

import matplotlib.pyplot as plt

from gaiaxpy.core.satellite import BANDS
from .plotter import Plotter


class MultiXpPlotter(Plotter):

    def _plot_multi_xp(self):
        show_legend = self.legend
        spectra_df = self.spectra
        spectra_class = self.spectra_class
        max_flux = 0
        fig, ax = plt.subplots(ncols=2, figsize=(16, 8))
        # Set titles
        ax[0].set_title(BANDS.bp.upper())
        ax[1].set_title(BANDS.rp.upper())
        ax[1].yaxis.set_label_position('right')
        ax[1].yaxis.tick_right()
        for spectrum in spectra_df.to_dict('records'):
            x, y, e = self._get_inputs(spectrum)
            max_flux = max(y) if max(y) > max_flux else max_flux
            axis = BANDS.index(spectrum['xp'].lower())
            ax[axis].plot(x, y, lw=2, alpha=0.95, label=spectrum['source_id'])
        for axis in ax:
            axis.set_ylim(0, 1.05 * max_flux)
            axis.set_xlabel(spectra_class.get_position_label())
            axis.set_ylabel(spectra_class.get_flux_label())
            if show_legend:
                handles, labels = plt.gca().get_legend_handles_labels()
                by_label = dict(zip(labels, handles))
                fig.subplots_adjust(top=0.95, bottom=0.05, left=0.05, right=0.85, wspace=0.020)
                fig.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1, 1), loc='upper right',
                           borderaxespad=0.1)

    def plot_fig(self):
        n_spectra = len(self.spectra)
        if self.show_plot and self.legend and n_spectra > self.max_spectra_on_multi:
            raise ValueError(f'The legend can only be shown for a list of spectra no longer than '
                             f'{self.max_spectra_on_multi} elements. Try setting legend to False or retry with a '
                             f'shorter list.')
        self._plot_multi_xp()
        if self.output_path or self.output_file:
            self._save_figure()
        if self.show_plot:
            plt.show()
        plt.close()



class DataFrameNumPyArrayReader(object):

    def __init__(self, content, array_columns):
        self.content = content
        self.array_columns = array_columns

    def read(self):
        df = self.content
        return df



from pathlib import Path
from typing import Union

from gaiaxpy.file_parser.parse_generic import GenericParser


class InverseBasesParser(GenericParser):
    """
    Parser for the inverse bases used in the external calibration.
    """

    def _parse_csv(self, csv_file: Union[Path, str], _array_columns=None, _matrix_columns=None):
        """
        Parse the input CSV file and store the result in a pandas DataFrame if it contains inverse bases.

        Args:
            csv_file (Path/str): Path to a CSV file.

        Returns:
            DataFrame: Pandas DataFrame populated with the content of the CSV file.
        """
        if _matrix_columns is None:
            _matrix_columns = []
        if _array_columns is None:
            _array_columns = ['inverseBasesCoefficients', 'transformationMatrix']
        return super()._parse_csv(csv_file, _array_columns=_array_columns, _matrix_columns=_matrix_columns)



import sys
import tempfile
from configparser import ConfigParser
from os import walk, urandom
from os.path import isdir, join
from re import match

from gaiaxpy.core.config import ADDITIONAL_SYSTEM_PREFIX

_CFG_FILE_PATH = join(tempfile.gettempdir(), urandom(24).hex())
_ADDITIONAL_SYSTEM_FILES_REGEX = r'[a-zA-Z0-9-_]+\.gaiaxpy_dr3_[a-zA-Z0-9-]+\.xml'


class GenCfg:

    def __init__(self, filters_dir: str, version: str):
        if not isdir(filters_dir):
            raise ValueError(f'{filters_dir} is not a path to a valid directory.')
        self.filters_dir = filters_dir
        self.version = version
        self.filter = 'system.gaiaXPy_dr3_version.xml'


def __get_file_names_recursively(dir_path: str, show_warning: bool = False) -> list:
    """
    Get all compliant file names recursively from a directory.

    Args:
        dir_path (str): Path to the directory where the files are located.

    Returns:
        list: List of compliant file names.

    Raises:
        ValueError: If no compliant files are found in the given directory.
    """
    if not isdir(dir_path):
        raise ValueError('Input path is not a valid directory.')
    all_files = [f for _, _, fn in walk(dir_path) for f in fn]
    compliant_files = [f for f in all_files if __file_name_is_compliant(f)]
    if len(compliant_files) == 0:
        raise ValueError('No filter files found in the given directory. Please check your files.')
    elif len(compliant_files) < len(all_files) and show_warning:
        message = 'Based on their names, some files in the directory do not correspond to filter files. They will be ignored.'
        print(f'UserWarning: {message}', file=sys.stderr)
    return compliant_files


def __file_name_is_compliant(file_name: str) -> bool:
    """
    Check if a file name is compliant with the naming convention.

    Args:
        file_name (str): Name of the file.

    Returns:
        bool: True if the file name is compliant, False otherwise.
    """
    return match(_ADDITIONAL_SYSTEM_FILES_REGEX, file_name.lower()) is not None


def create_config(systems_path: str = None, config_file: str = None):
    """
    Creates a configuration file with the provided `systems_path` and `config_file` parameters.
    If `systems_path` is not provided, it will prompt the user to enter it.
    The configuration file contains the path to the additional systems directory and the version of the additional
    systems. If more than one version of the additional systems is detected, a ValueError will be raised.

    Args:
        systems_path (str): The path to the additional systems' directory. If not provided, it will be prompted.
        config_file (str): The path to the configuration file. If not provided, the default path will be used.
    """
    quotes = ["'", '"']
    if not systems_path:
        systems_path = input('Please enter the path to the systems directory: ')
    systems_path = systems_path[1:-1] if systems_path[0] in quotes and systems_path[-1] in quotes else systems_path
    # Get filters version
    files = __get_file_names_recursively(systems_path, show_warning=True)
    version = [f.split('.')[1].split('_')[-1] for f in files]
    if len(set(version)) != 1:
        raise ValueError('More than one version detected in the additional systems. This is currently not allowed.')
    elif len(set(version)) == 1:
        version = version[0]
    cfg_details = GenCfg(systems_path, version)
    __write_config(cfg_details, config_file)


def __write_config(cfg_details, config_file: str = None):
    """
    Write the configuration details to the config file.

    Args:
        cfg_details: Object with the configuration details.
        config_file (str): Path to the config file. Defaults to None.
    """
    config_file = _CFG_FILE_PATH if not config_file else config_file
    config = ConfigParser()
    config['filter'] = vars(cfg_details)
    with open(config_file, 'w') as cf:
        config.write(cf)
    print(f'Loading systems... Additional systems version is: {cfg_details.version}.')


def load_config(config_file: str = None):
    """
    Load the configuration file and return a ConfigParser object containing the parsed data.

    Args:
        config_file (str): The path to the configuration file. Defaults to None, which uses the default configuration
            file.

    Returns:
        ConfigParser: A ConfigParser object containing the parsed configuration data.
    """
    config_file = _CFG_FILE_PATH if not config_file else config_file
    config = ConfigParser()
    with open(config_file) as f:
        config.read_file(f)
    return config


def get_additional_filters_path(config_file: str = None):
    """
    Return the path to the directory containing the additional filters.

    Args:
        config_file (str): The path to the configuration file. If no path is given, the default configuration file will
            be used.

    Returns:
        str: The path to the directory containing the additional filters.
    """
    try:
        config = load_config(config_file)
        return config['filter']['filters_dir']
    except (IOError, KeyError):
        return None


def contains_filter_key(config_file: str = None):
    """
    Determine if the configuration file contains the key for the additional filters' directory.

    Args:
        config_file (str): The path to the configuration file. If no path is given, the default configuration file will
            be used.

    Returns:
        bool: True if the configuration file contains the key for the additional filters' directory, False otherwise.
    """
    config = load_config(config_file)
    try:
        k = bool(config['filter']['filters_dir'])
    except (IOError, KeyError):
        return False
    return k


def get_additional_filters_names(config_file: str = None):
    """
    Returns the names of the additional filters.

    Args:
        config_file (str): The path to the configuration file. If no path is given, the default configuration file will
            be used.

    Returns:
        list: The names of the additional filters.
    """
    filters_path = get_additional_filters_path(config_file)
    additional_system_names = []
    if filters_path:
        filenames = __get_file_names_recursively(filters_path)
        additional_system_names = [f"{ADDITIONAL_SYSTEM_PREFIX}_{f.split('.')[0]}" for f in filenames]
    return additional_system_names



from configparser import ConfigParser
from os.path import abspath, dirname, join

config_path = abspath(dirname(__file__))
filters_path = join(config_path, 'filters')

config_ini_file = join(config_path, 'config.ini')
configparser = ConfigParser()
configparser.read(config_ini_file)
hermite_bases_file = join(config_path, configparser.get('converter', 'hermite_bases'))
spline_bases_file = join(config_path, configparser.get('converter', 'spline_bases'))

correction_tables_path = join(config_path, 'correction_tables')



"""
standardised_photometric_system.py
====================================
Module to represent a standardised photometric system.
"""

from .internal_photometric_system import InternalPhotometricSystem


class StandardisedPhotometricSystem(InternalPhotometricSystem):

    def __init__(self, name, config_file=None):
        """
        A photometric system is defined by the set of bands available.

        Args:
            name (str): Name of the PhotometricSystem
        """
        super().__init__(name, config_file)
        self._load_offset_from_xml()

    def _correct_flux(self, flux):
        flux_corr = flux + self.offsets
        return flux_corr

    def _correct_error(self, flux, error):
        eef = (flux + self.offsets) / flux
        eef = abs(eef)
        flux_error_corr = error * eef
        return flux_error_corr



"""
nature.py
====================================
Module to hold "natural" constants.
"""

# Constants extracted from Gaia Parameter Database (Nature section)
C = 2.99792458E8
PLANCK = 6.62607004E-34



"""
single.py
====================================
Module to plot a single spectrum, either absolute or XP.
"""

import matplotlib.pyplot as plt

from .plotter import Plotter


class SinglePlotter(Plotter):

    def _plot_single(self, spectrum):
        spectra_class = self.spectra_class
        source_id = spectrum['source_id']
        fig, ax = plt.subplots(figsize=(16, 9))
        x, y, e = self._get_inputs(spectrum)
        ax.plot(x, y, lw=2, alpha=0.95, label=source_id)
        ax.fill_between(x, y - e, y + e, alpha=0.2)
        ax.set_title('{}'.format(source_id))
        ax.set_xlabel(spectra_class.get_position_label())
        ax.set_ylabel(spectra_class.get_flux_label())
        plt.tight_layout()

    def plot_fig(self):
        n_spectra = len(self.spectra)
        if n_spectra > self.max_spectra_on_single:
            raise ValueError(
                f'Spectra list is too long. This functionality can only show up to {self.max_spectra_on_single} '
                f'single plots. Try saving the plots without showing them using the option output_path.')
        records = self.spectra.to_dict('records')
        is_len_1 = len(records) == 1
        for index, spectrum in enumerate(records):
            self._plot_single(spectrum)
            if self.save_file:
                index = index if not is_len_1 else None
                self._save_figure(index=index)
            if self.show_plot:
                plt.show()
        plt.close()



"""
cholesky.py
====================================
Module that implements the Cholesky functionality.
"""
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from numpy import diag, dot, identity
from scipy.linalg import cholesky, solve_triangular

from gaiaxpy.core.generic_functions import parse_band
from gaiaxpy.core.satellite import BANDS
from gaiaxpy.input_reader.input_reader import InputReader


def __get_dot_product(_L_inv: np.ndarray) -> Optional[np.ndarray]:
    """
    Calculate the dot product of the transpose of L_inv with itself.

    Args:
        _L_inv (ndarray): A 2D numpy array.

    Returns:
        ndarray: The dot product of the transpose of L_inv with itself. None: If L_inv does not have the attribute `T`
            (transpose).
    """
    try:
        return dot(_L_inv.T, _L_inv)
    except AttributeError:
        return None


def __output_list_to_df(parsed_input_data: pd.DataFrame, bands_output: list, output_columns: list) -> pd.DataFrame:
    """
    Convert a list of output data into a pandas DataFrame.

    Args:
        parsed_input_data (pd.Dataframe): Parsed initial data.
        bands_output (list): A list of output data, where each element is a list of values for a specific band.
        output_columns (list): A list of strings representing the column names for the output DataFrame.

    Returns:
        pd.DataFrame: A DataFrame containing the source ID and the output data for each band.
    """
    output_list = [parsed_input_data['source_id']]
    for element in bands_output:
        output_list.append(element)
    return pd.DataFrame(zip(*output_list), columns=output_columns)


def get_inverse_square_root_covariance_matrix(input_object: Union[list, Path, pd.DataFrame, str],
                                              band: Optional[str] = None):
    """
    Compute the inverse square root covariance matrix.

    Args:
        input_object (list/Path/pd.DataFrame/str): Path to the file containing the mean spectra as downloaded from the
            Archive in their continuous representation, a pandas DataFrame, a list of sources ids (string or long), or
            an ADQL query.
        band (str): Chosen band: 'bp' or 'rp'. If no band is passed, the function will compute the inverse square root
            covariance for both 'bp' and 'rp'.

    Returns:
        DataFrame or ndarray: DataFrame containing the source IDs and the output inverse square root covariance matrices
            for the sources in the input object if it contains more than one source or no band is passed to the
            function.
            The function will return a ndarray (of shape (55, 55)) if there is only one source ID in the input data
            and a single band is selected.
    """
    if band is not None:
        band = parse_band(band)
    parsed_input_data, extension = InputReader(input_object, get_inverse_square_root_covariance_matrix, False).read()
    if band is None:
        bands_to_process = BANDS
        output_columns = ['source_id', 'bp_inverse_square_root_covariance_matrix',
                          'rp_inverse_square_root_covariance_matrix']
    else:
        bands_to_process = [band]
        output_columns = ['source_id', f'{band}_inverse_square_root_covariance_matrix']
    bands_output = []
    for b in bands_to_process:
        xp_errors = parsed_input_data[f'{b}_coefficient_errors'] / parsed_input_data[f'{b}_standard_deviation']
        xp_correlation_matrix = parsed_input_data[f'{b}_coefficient_correlations']
        band_inv_iterable = map(_get_inverse_square_root_covariance_matrix_aux, xp_errors, xp_correlation_matrix)
        bands_output.append(band_inv_iterable)
    output_df = __output_list_to_df(parsed_input_data, bands_output, output_columns)
    if len(bands_to_process) == 1 and len(output_df) == 1:
        return output_df[f'{band}_inverse_square_root_covariance_matrix'].iloc[0]
    else:
        return output_df


def _get_inverse_square_root_covariance_matrix_aux(xp_errors: np.ndarray, xp_correlation_matrix: np.ndarray) -> \
        Optional[np.ndarray]:
    """
    Calculate the inverse square root of the covariance matrix.

    Args:
        xp_errors (ndarray): A numpy array representing the measurement errors.
        xp_correlation_matrix (ndarray): A numpy array representing the correlation matrix.

    Returns:
        ndarray: The inverse square root of the covariance matrix. None: If the Cholesky decomposition of the
            correlation matrix fails.
    """
    try:
        _L = cholesky(xp_correlation_matrix, lower=True)
        # Invert lower triangular matrix
        _L_inv = solve_triangular(_L, identity(len(_L)), lower=True)
        # Matrix of inverse errors
        _E_inv = diag(1.0 / xp_errors)
        return dot(_L_inv, _E_inv)
    except ValueError:
        return None


def get_inverse_covariance_matrix(input_object: Union[list, Path, str], band: Optional[str] = None):
    """
    Compute the inverse covariance matrix.

    Args:
        input_object (object): Path to the file containing the mean spectra as downloaded from the archive in their
            continuous representation, a pandas DataFrame, a list of sources ids (string or long), or an ADQL query.
        band (str): Chosen band: 'bp' or 'rp'. If no band is passed, the function will compute the inverse covariance
            for both 'bp' and 'rp'.

    Returns:
        DataFrame or ndarray: DataFrame containing the source IDs and the output inverse covariance matrices for the
            sources in the input object if it contains more than one source or no band is passed to the function.
            The function will return a ndarray (of shape (55, 55)) if there is only one source ID in the input data
            and a single band is selected.
    """
    band = band if band is None else parse_band(band)
    parsed_input_data, extension = InputReader(input_object, get_inverse_covariance_matrix, False).read()
    bands_output = []
    if band is None:
        bands_to_process = BANDS
        output_columns = ['source_id', 'bp_inverse_covariance', 'rp_inverse_covariance']
    else:
        bands_to_process = [band]
        output_columns = ['source_id', f'{band}_inverse_covariance']
    for b in bands_to_process:
        # The formal errors need to be scaled by the inverse standard deviation.
        xp_errors = parsed_input_data[f'{b}_coefficient_errors'] / parsed_input_data[f'{b}_standard_deviation']
        xp_correlation_matrix = parsed_input_data[f'{b}_coefficient_correlations']
        _L_inv_iterable = map(_get_inverse_square_root_covariance_matrix_aux, xp_errors, xp_correlation_matrix)
        band_output = map(__get_dot_product, _L_inv_iterable)
        bands_output.append(band_output)
    output_df = __output_list_to_df(parsed_input_data, bands_output, output_columns)
    if len(bands_to_process) == 1 and len(output_df) == 1:
        return output_df[f'{band}_inverse_covariance'].iloc[0]
    else:
        return output_df


def get_chi2(_L_inv: np.ndarray, residuals: np.ndarray) -> np.ndarray:
    """
    Compute chi-squared (chi2) from given inverse Cholesky of the covariance matrix (L^-1) and a residual vector
    (r = data - model). This function defines x = L^-1 * r such that chi2 = |x|^2, which guarantees that chi2 >= 0.

    Args:
        _L_inv (ndarray): Inverse square root of the covariance, as computed from the function
            get_inverse_square_root_covariance_matrix.
        residuals (ndarray): Difference between the observed coefficient vector and some model prediction of it.

    Returns:
        float: Chi-squared value.
    """
    if _L_inv is None or residuals is None:
        raise ValueError('Input parameters cannot be None.')
    if _L_inv.shape != (55, 55):
        raise ValueError('Inverse covariance matrix shape must be (55, 55).')
    if residuals.shape != (55,):
        raise ValueError('Residuals shape must be (55,).')
    x = dot(_L_inv, residuals)
    return dot(x.T, x)



"""
cast.py
====================================
Module to cast the data after parsing.
"""
import numpy as np
from numpy.ma import MaskError, getdata
from pandas.errors import IntCastingNaNError

# Fields of all formats including AVRO.
__type_map = {'source_id': 'Int64', 'solution_id': 'Int64',
              'rp_n_parameters': 'Int16', 'bp_n_parameters': 'Int16',
              'rp_n_rejected_measurements': 'Int64', 'bp_n_rejected_measurements': 'Int64',
              'rp_n_measurements': 'Int64', 'bp_n_measurements': 'Int64',
              'rp_standard_deviation': 'Float64', 'bp_standard_deviation': 'Float64',
              'rp_num_of_transits': 'Int64', 'bp_num_of_transits': 'Int64',
              'rp_num_of_blended_transits': 'Int64', 'bp_num_of_blended_transits': 'Int64',
              'rp_num_of_contaminated_transits': 'Int64', 'bp_num_of_contaminated_transits': 'Int64',
              'rp_coefficients': 'O', 'bp_coefficients': 'O',
              'rp_coefficient_covariances': 'O', 'bp_coefficient_covariances': 'O',
              'rp_degrees_of_freedom': 'Int64', 'bp_degrees_of_freedom': 'Int64',
              'rp_n_relevant_bases': 'Int64', 'bp_n_relevant_bases': 'Int64',
              'rp_basis_function_id': 'Int64', 'bp_basis_function_id': 'Int64',
              'rp_chi_squared': 'Float64', 'bp_chi_squared': 'Float64',
              'rp_coefficient_errors': 'O', 'bp_coefficient_errors': 'O',
              'rp_coefficient_correlations': 'O', 'bp_coefficient_correlations': 'O',
              'rp_relative_shrinking': 'Float64', 'bp_relative_shrinking': 'Float64'}


def __replace_masked_array(value):
    if (isinstance(value, np.ma.core.MaskedArray) and getdata(value).size == 0) or (isinstance(value, float) and
                                                                                    value == 0.0):
        return np.array([])
    elif isinstance(value, np.ma.core.MaskedArray):
        return getdata(value)
    else:
        return value


def __replace_masked_constant(value):
    if isinstance(value, np.ma.core.MaskedConstant):
        return np.nan if value.item() == 0.0 else value
    return value


def _cast(df):
    """
    Cast types to the defined ones to standardise the different input formats.

    Args:
        df (DataFrame): a DataFrame with parsed data from input files.
    """
    # flake8: noqa
    for column in ['bp_n_parameters', 'bp_basis_function_id']:
        if column in df.columns:
            df[column] = df[column].apply(lambda row: __replace_masked_constant(row))
    for column, type_value in __type_map.items():
        try:
            if type_value == 'O':
                df[column] = df[column].apply(lambda x: __replace_masked_array(x))
            else:
                df[column] = df[column].astype(type_value)
        except (TypeError, IntCastingNaNError):
            df[column] = df[column].astype(type_value.title())
        except KeyError:
            continue  # Not every key is available in every case
        except ValueError as err:
            if df[column].isna().any():
                pass  # There can be nan values, do nothing with them at this step
            else:
                # This is an actual error
                raise ValueError(f'Error casting input data: {err}. Value {df[column]} in column {column} cannot be'
                                 f' casted to type  {type_value}.')
        except MaskError:
            continue
    return df



"""
multi_absolute.py
====================================
Module to plot multiple absolute spectra.
"""

import matplotlib.pyplot as plt

from .plotter import Plotter


class MultiAbsolutePlotter(Plotter):

    def __plot_multi_absolute(self):
        spectra_class = self.spectra_class
        spectra_df = self.spectra
        fig, ax = plt.subplots(figsize=(16, 9))
        for spectrum in spectra_df.to_dict('records'):
            x, y, e = self._get_inputs(spectrum)
            ax.plot(x, y, lw=2, alpha=0.95, label=spectrum['source_id'])
        ax.set_xlabel(spectra_class.get_position_label())
        ax.set_ylabel(spectra_class.get_flux_label())
        fig.tight_layout()
        if self.legend:
            by_label = dict()
            for ax in fig.axes:
                handles, labels = plt.gca().get_legend_handles_labels()
                by_label = dict(zip(labels, handles))
            fig.subplots_adjust(top=0.95, bottom=0.05, left=0.05, right=0.85, wspace=0.020)
            fig.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1, 1), loc='upper right', borderaxespad=1.0)
        return fig, ax

    def plot_fig(self):
        n_spectra = len(self.spectra)
        if self.show_plot and self.legend and n_spectra > self.max_spectra_on_multi:
            raise ValueError(f'The legend can only be shown for a list of spectra no longer than '
                             f'{self.max_spectra_on_multi} elements. Try setting legend to False or retry with a '
                             f'shorter list.')
        self.__plot_multi_absolute()
        if self.save_file:
            self._save_figure()
        if self.show_plot:
            plt.show()
        plt.close()



import re

from gaiaxpy.core.server import data_release, gaia_server
from gaiaxpy.core.version import __version__
from .archive_reader import ArchiveReader
from .dataframe_reader import DataFrameReader
from ..core.custom_errors import SelectorNotImplementedError

not_supported_functions = ['apply_colour_equation']


class QueryReader(ArchiveReader):

    def __init__(self, content, function, truncation, user=None, password=None, additional_columns=None, selector=None,
                 disable_info=False):
        if additional_columns is None:
            additional_columns = dict()
        if selector is not None:
            raise SelectorNotImplementedError('Query')
        self.content = content
        super(QueryReader, self).__init__(function, truncation, user, password, additional_columns=additional_columns,
                                          disable_info=disable_info)

    @staticmethod
    def get_srcids(_table):
        """
        Get the name of the source ID column from an Astropy Table.

        Args:
            _table (Table): A table containing columns.

        Returns:
            str: The name of the source ID column.

        Raises:
            ValueError: If the source ID column is not found in the DataFrame.
            ValueError: If the index of the source ID column is not a string.
        """
        try:
            stripped_pairs = ((re.sub(r'[-_ ]', '', c.lower()), c) for c in _table.columns.keys())
            sid_col = next(original for stripped, original in stripped_pairs if stripped in ['sourceid', 'srcid'])
        except StopIteration:
            raise ValueError('Source ID column not found in query result.')
        if not isinstance(sid_col, str):
            raise ValueError(
                f'Index of source ID column should be a string, but is {type(sid_col).__name__}: {sid_col}.')
        return _table[sid_col]

    @staticmethod
    def _add_marker(query, comment):
        def __remove_comments(_query: str) -> str:
            single_line_comment = re.compile(r'--.*$', re.MULTILINE)
            multi_line_comment = re.compile(r'/\*.*?\*/', re.DOTALL)
            query_no_single_line_comments = single_line_comment.sub('', _query)  # Remove single-line comments
            return multi_line_comment.sub('', query_no_single_line_comments)  # Remove multi-line comments

        query = __remove_comments(query)

        if comment:
            insensitive_select = re.compile(re.escape('select'), re.IGNORECASE)
            query = insensitive_select.sub(f'select --{comment} \n', query)
        return query

    def read(self, _data_release=data_release, _comment=f'This query was launched from within GaiaXPy {__version__}'):
        # Import only when read is called, as this module is dependent on the Archive’s availability
        from astroquery.gaia import GaiaClass

        query = self.content
        query = self._add_marker(query, _comment)
        function_name = self.function.__name__
        if function_name in not_supported_functions:
            raise ValueError(f'Function {function_name} does not accept ADQL queries.')
        # Connect to geapre
        gaia = GaiaClass(gaia_tap_server=gaia_server, gaia_data_server=gaia_server)
        self._login(gaia)
        # ADQL query
        if not self.disable_info:
            self.show_info_msg()
        job = gaia.launch_job_async(query, dump_to_file=False)
        query_result = job.get_results()
        result = gaia.load_data(ids=self.get_srcids(query_result), format='csv', data_release=_data_release,
                                data_structure='raw', retrieval_type='XP_CONTINUOUS', avoid_datatype_check=True)
        try:
            continuous_key = [key for key in result.keys() if 'continuous' in key.lower()][0]
            data = result[continuous_key][0].to_pandas()
        except KeyError:
            raise ValueError('No continuous BP/RP data found for the requested query.')
        if not self.disable_info:
            self.show_info_msg(done=True)
        return DataFrameReader(data, function_name, self.truncation, additional_columns=self.additional_columns,
                               disable_info=True).read()



"""
sampled_basis_functions.py
====================================
Module to represent a set of basis functions evaluated on a grid.
"""

import functools
import math

import numpy as np
from scipy.interpolate import BSpline
from scipy.special import eval_hermite, gamma

from gaiaxpy.core import nature, satellite

sqrt_4_pi = np.pi ** (-0.25)


class SampledBasisFunctions(object):
    """
    Evaluation of a set of basis functions on a user-defined grid.
    """

    def __init__(self, sampling_grid, design_matrix=None):
        """
        Initialise a sampled basis functions object.

        Args:
            sampling_grid (ndarray): 1D array of positions where the bases need to be evaluated.
            design_matrix (ndarray): 2D array containing the evaluation of each basis at all positions in the sampling
                grid.
        """
        self.sampling_grid = sampling_grid
        self.design_matrix = design_matrix

    @classmethod
    def from_external_instrument_model(cls, sampling, weights, external_instrument_model):
        """
        Instantiate an object starting from a sampling grid, an array of weights and the external calibration instrument
            model.

        Args:
            sampling (ndarray): 1D array of positions where the bases need to be evaluated.
            weights (ndarray): 1D array containing the weights to be applied at each element in the sampling grid. These
                are simply used to define where in the sampling grid some contribution is expected. Where the weight is
                0, the bases will not be evaluated.
            external_instrument_model (obj): external calibration instrument model. This object contains information on
                the dispersion, response and inverse bases.

        Returns:
            SampledBasisFunctions: An instance of this class.
        """
        n_samples = len(sampling)
        scale = ((external_instrument_model.bases['normRangeMax'] -
                  external_instrument_model.bases['normRangeMin']) /
                 (external_instrument_model.bases['pwlRangeMax'] -
                  external_instrument_model.bases['pwlRangeMin']))
        offset = (external_instrument_model.bases['normRangeMin'] -
                  external_instrument_model.bases['pwlRangeMin'] * scale)

        sampling_pwl = external_instrument_model.wl_to_pwl(sampling)
        rescaled_pwl = (sampling_pwl * scale) + offset

        bases_transformation = external_instrument_model.bases['transformationMatrix']
        evaluated_hermite_bases = np.array([_evaluate_hermite_function(n_h, pos, weight) for pos, weight in
                                            zip(rescaled_pwl, weights) for n_h in np.arange(
                int(external_instrument_model.bases['nInverseBasesCoefficients']))]).reshape(
            n_samples, int(external_instrument_model.bases['nInverseBasesCoefficients']))
        _design_matrix = external_instrument_model.bases['inverseBasesCoefficients'] @ evaluated_hermite_bases.T
        transformed_design_matrix = bases_transformation @ _design_matrix

        hc = 1.e9 * nature.C * nature.PLANCK

        def compute_norm(wl):
            r = external_instrument_model.get_response(wl)
            if r > 0:
                return hc / (satellite.TELESCOPE_PUPIL_AREA * r * wl)
            else:
                return 0.0

        norm = np.array([compute_norm(wl) for wl in sampling])
        design_matrix = np.array([transformed_design_matrix[i] * norm for i in
                                  np.arange(external_instrument_model.bases['nBases'])])

        return cls(sampling, design_matrix=design_matrix)

    @classmethod
    def from_config(cls, sampling, bases_config):
        """
        Instantiate an object starting from a sampling grid and the configuration for the basis functions.

        Args:
            sampling (ndarray): 1D array of positions where the bases need to be evaluated.
            bases_config (DataFrame): The configuration of the set of bases loaded into a DataFrame.

        Returns:
            object: An instance of this class.
        """
        design_matrix = populate_design_matrix(sampling, bases_config)
        return cls(sampling, design_matrix=design_matrix)

    @classmethod
    def from_design_matrix(cls, sampling, design_matrix):
        """
        Instantiate an object starting from a sampling grid and the design matrix.

        Args:
            sampling (ndarray): 1D array of positions where the bases need to be evaluated.
            design_matrix (ndarray): 2D array containing the evaluation of each basis at all positions in the sampling
                grid.

        Returns:
            object: An instance of this class.
        """
        return cls(sampling, design_matrix=design_matrix)

    def get_design_matrix(self):
        return self.design_matrix

    def get_sampling_grid(self):
        return self.sampling_grid


def _evaluate_hermite_function(n, x, w):
    return _hermite_function(n, x) if w > 0 else 0


@functools.lru_cache(maxsize=128)
def _hermite_function(n, x):
    if n == 0:
        return sqrt_4_pi * np.exp(-x ** 2. / 2.)
    elif n == 1:
        return sqrt_4_pi * np.exp(-x ** 2. / 2.) * np.sqrt(2.) * x
    c1 = np.sqrt(2. / n) * x
    c2 = -np.sqrt((n - 1) / n)
    return c1 * _hermite_function(n - 1, x) + c2 * _hermite_function(n - 2, x)


def populate_design_matrix(sampling_grid, bases_config):
    def __psi(n, x):
        return (1.0 / np.sqrt(math.pow(2, n) * gamma(n + 1) * np.sqrt(np.pi)) * np.exp(-x ** 2 / 2.0) *
                eval_hermite(n, x))

    n_samples = len(sampling_grid)
    bc_columns = bases_config.columns
    if 'knots' not in bc_columns and 'transformedSetDimension' in bc_columns:  # Hermite
        normalised_range_lower, normalised_range_upper = bases_config['normalizedRange'].iloc(0)[0]
        range_lower, range_upper = bases_config['range'].iloc(0)[0]
        scale = (normalised_range_upper - normalised_range_lower) / (range_upper - range_lower)
        offset = normalised_range_lower - range_lower * scale
        rescaled_pwl = (sampling_grid * scale) + offset
        dimension = int(bases_config['dimension'].iloc[0])
        transformed_set_dimension = int(bases_config['transformedSetDimension'].iloc[0])
        bases_transformation = bases_config['transformationMatrix'].iloc(0)[0].reshape(dimension,
                                                                                       transformed_set_dimension)
        design_matrix = np.array([__psi(n_h, pos) for pos in rescaled_pwl for n_h in np.arange(dimension)]).reshape(
            n_samples, dimension)
        return bases_transformation @ design_matrix.T
    elif 'knots' in bc_columns:  # Spline
        if len(bases_config) != 1:
            raise ValueError('Only one row should be accepted at a time.')
        knots = bases_config['knots'].iloc[0]
        n_knots = len(knots)
        order = bases_config['order'].iloc[0]
        degree = order - 1
        n_bases = n_knots - order
        if 'transformationMatrix' in bases_config.__dir__():
            transformation_matrix = np.array(bases_config.transformationMatrix.values[0])
            ts_dim = bases_config.transformedSetDimension.values[0]
            bases_transformation = transformation_matrix.reshape(ts_dim, n_bases)
        else:
            bases_transformation = np.identity(n_bases)
        design_matrix = np.zeros((n_bases, n_samples))

        for basis_id in np.arange(n_bases):  # Evaluate
            c = np.zeros(n_knots)
            c[basis_id] = 1.0
            basis = BSpline(knots, c, degree)
            design_matrix[basis_id] = np.array([basis(pos) for pos in sampling_grid])
        return bases_transformation.dot(design_matrix)
    else:
        raise ValueError('Design matrix cannot be populated from the given configuration.')



"""
utils.py
====================================
Module containing auxiliary functions of the parsers.
"""

import operator
from functools import reduce


def _get_from_dict(dictionary, _map):
    try:
        return reduce(operator.getitem, _map, dictionary)
    except TypeError:
        return float('NaN')
    except KeyError:
        raise KeyError(f'Element {_map} not found in AVRO dictionary. Is it an actual field in the input file?')


# This dictionary contains the mapping from the usual CSV fields to the AVRO fields.
_csv_to_avro_map = {'source_id': ['sourceId'],
                    'rp_n_rejected_measurements': ['rpSpec', 'solution', 'numberOfRejectedMeasurements'],
                    'rp_chi_squared': ['rpSpec', 'solution', 'chiSquared'],
                    'rp_degrees_of_freedom': ['rpSpec', 'solution', 'degreesOfFreedom'],
                    'rp_n_parameters': ['rpSpec', 'solution', 'numberOfParameters'],
                    'rp_standard_deviation': ['rpSpec', 'solution', 'standardDeviation'],
                    'rp_n_measurements': ['rpSpec', 'solution', 'numberOfMeasurements'],
                    'rp_n_relevant_bases': ['rpSpec', 'NRelevantBases'],
                    'rp_basis_function_id': ['rpSpec', 'basisFunctionSetDefId'],
                    'rp_num_of_transits': ['rpNumOfTransits'],
                    'bp_n_rejected_measurements': ['bpSpec', 'solution', 'numberOfRejectedMeasurements'],
                    'bp_chi_squared': ['bpSpec', 'solution', 'chiSquared'],
                    'bp_degrees_of_freedom': ['bpSpec', 'solution', 'degreesOfFreedom'],
                    'bp_n_parameters': ['bpSpec', 'solution', 'numberOfParameters'],
                    'bp_standard_deviation': ['bpSpec', 'solution', 'standardDeviation'],
                    'bp_n_measurements': ['bpSpec', 'solution', 'numberOfMeasurements'],
                    'bp_n_relevant_bases': ['bpSpec', 'NRelevantBases'],
                    'bp_basis_function_id': ['bpSpec', 'basisFunctionSetDefId'],
                    'rp_num_of_blended_transits': ['rpNumOfBlendedTransits'],
                    'bp_num_of_transits': ['bpNumOfTransits'],
                    'bp_num_of_contaminated_transits': ['bpNumOfContaminatedTransits'],
                    'rp_num_of_contaminated_transits': ['rpNumOfContaminatedTransits'],
                    'bp_num_of_blended_transits': ['bpNumOfBlendedTransits'],
                    'solution_id': ['solutionId'],
                    'rp_coefficient_covariances': ['rpSpec', 'solution', 'covariance'],
                    'rp_coefficients': ['rpSpec', 'solution', 'parameters'],
                    'bp_coefficient_covariances': ['bpSpec', 'solution', 'covariance'],
                    'bp_coefficients': ['bpSpec', 'solution', 'parameters']}



"""
internal_photometric_system.py
====================================
Module for the parent class of the standardised and regular photometric systems.
"""
import re
from configparser import ConfigParser
from glob import glob
from os import remove
from os.path import exists, split, join

from gaiaxpy.config.paths import config_ini_file
from gaiaxpy.core.config import (get_filter_version_from_config, replace_file_name, get_file_path,
                                 ADDITIONAL_SYSTEM_PREFIX)
from gaiaxpy.core.generic_functions import _get_system_label
from gaiaxpy.core.satellite import BANDS
from gaiaxpy.core.version import __version__
from gaiaxpy.core.xml_utils import get_file_root, parse_array, get_array_text, get_xp_sampling_matrix, get_xp_merge
from .config import _CFG_FILE_PATH, _ADDITIONAL_SYSTEM_FILES_REGEX


class InternalPhotometricSystem(object):

    def __init__(self, name: str, config_file: str = None, bp_model: str = 'v375wi', rp_model: str = 'v142r'):
        self.label = _get_system_label(name)
        self.version = None
        self.__set_version(config_file)
        config_file = config_ini_file if not config_file else config_file
        self.config_file = config_file
        self.filter_file = None
        self._set_file(bp_model=bp_model, rp_model=rp_model)
        self.bands = None
        self.zero_points = None
        self._load_xpzeropoint_from_xml()
        self.offsets = None
        self._load_offset_from_xml()
        self.name = name

    def set_bands(self, bands):
        """
        Set the bands of the photometric system.

        Args:
            bands (list): List of bands in this photometric system.
        """
        self.bands = list(bands)

    def get_bands(self):
        """
        Get the bands of the photometric system.

        Returns:
            list of str: List of bands.
        """
        return self.bands

    def set_offsets(self, offsets):
        self.offsets = offsets

    def get_offsets(self):
        return self.offsets

    def get_system_label(self):
        """
        Get the label of the photometric system.

        Returns:
            str: A short description of the photometric system.
        """
        return self.label

    def set_zero_points(self, zero_points):
        """
        Set the zero-points needed to convert the Gaia fluxes in the bands defining this photometric system to
            magnitudes.

        Args:
            zero_points (nparray): 1D array containing the zero-point for each of the bands in this photometric system.
        """
        self.zero_points = zero_points

    def __set_version(self, config_file):
        if not config_file:
            self.version = __version__
        else:
            _config_parser = ConfigParser()
            _config_parser.read(config_file)
            self.version = get_filter_version_from_config(_config_parser)

    def get_zero_points(self):
        """
        Get the zero-points of the photometric system.

        Returns:
            ndarray: 1D array containing the zero-points for all bands in
            this photometric system.
        """
        return self.zero_points

    def _correct_flux(self, flux):
        raise ValueError('Method not implemented in parent class.')

    def _correct_error(self, flux, error):
        raise ValueError('Method not implemented in parent class.')

    def _set_file(self, bp_model, rp_model):
        """
        Get the file path corresponding to the given label and key.

        Args:
            bp_model (str): BP model.
            rp_model (str): RP model.

        Returns:
            str: Path of a file.
        """

        def _extract_matches(_actual_path, _system_name):
            def __is_built_in_name(f_name):
                return f_name.startswith('XpFilter')

            def __is_exact_match(f_name, sys_name):
                return f_name.startswith(sys_name) and f'{sys_name}.' in f_name

            def __matches_additional(f_name, _pattern):
                return _pattern.match(f_name)

            __split_paths = [split(p) for p in _actual_path]
            __paths, __filenames = zip(*__split_paths)
            pattern = re.compile(_ADDITIONAL_SYSTEM_FILES_REGEX, re.IGNORECASE)
            if all(__is_built_in_name(f) for f in __filenames):  # If system is built-in
                return _actual_path
            return [join(p, fn) for p, fn in __split_paths if not __is_built_in_name(fn) and
                    __matches_additional(fn, pattern) and __is_exact_match(fn, _system_name)]  # If is additional

        def _validate_path(_actual_path):
            if len(actual_path) == 0:
                raise ValueError('Filter file not found in given path.')
            elif len(actual_path) > 1:
                # Remove configuration file if it exists to avoid issues when reloading
                if exists(_CFG_FILE_PATH):
                    remove(_CFG_FILE_PATH)
                raise ValueError(f'More than one system named {self.label.replace(f"{ADDITIONAL_SYSTEM_PREFIX}_", "")}'
                                 f' were found. System names in the given directory should be unique. Operation aborted.')

        file_name = replace_file_name(self.config_file, 'filter', 'filter', bp_model, rp_model, self.label)
        system_name = file_name.split('.')[0]
        file_path = get_file_path(self.config_file)
        # Search file in file path to obtain the actual path
        actual_path = glob(file_path + f"/**/{system_name}*.xml", recursive=True)
        actual_path = _extract_matches(actual_path, system_name)
        _validate_path(actual_path)
        self.filter_file = actual_path[0]

    def _load_offset_from_xml(self):
        """
        Load the offset of a standard photometric system from the filter XML file.

        Returns:
            ndarray: Array of offsets.
        """
        x_root = get_file_root(self.filter_file)
        self.offsets = parse_array(x_root, 'fluxBias')

    def _load_xpzeropoint_from_xml(self):
        """
        Load the zero-points for each band from the filter XML file.

        Returns:
            ndarray: Array of zero-points.
        """
        x_root = get_file_root(self.filter_file)
        self.zero_points = parse_array(x_root, 'zeropoints')
        self.bands, _ = get_array_text(x_root, 'bands')

    def load_xpsampling_from_xml(self):
        """
        Load the XpSampling table from the XML filter file.

        Returns:
            dict: A dictionary containing the XpSampling table with one entry for BP and one for RP.
        """
        x_root = get_file_root(self.filter_file)
        _, n_bands = get_array_text(x_root, 'bands')

        bp_sampling = get_xp_sampling_matrix(x_root, 'bp', n_bands)
        rp_sampling = get_xp_sampling_matrix(x_root, 'rp', n_bands)

        xp_sampling = dict(zip(BANDS, [bp_sampling, rp_sampling]))
        return xp_sampling

    def load_xpmerge_from_xml(self):
        """
        Load the XpMerge table from the filter XML file.

        Returns:
            ndarray: Array containing the sampling grid values.
            dict: A dictionary containing the XpMerge table with one entry for BP and one for RP.
        """
        x_root = get_file_root(self.filter_file)
        sampling_grid, bp_merge, rp_merge = get_xp_merge(x_root)
        return sampling_grid, dict(zip(BANDS, [bp_merge, rp_merge]))



"""
external_instrument_model.py
====================================
Module for handling the various components of the external calibration instrument model.
These are dispersion function, instrument response and set of inverse bases.
"""

import numpy as np
import pandas as pd
from scipy import interpolate

from gaiaxpy.file_parser.parse_inverse import InverseBasesParser


class ExternalInstrumentModel(object):
    """
    External calibration instrument model.
    """

    def __init__(self, dispersion: dict, response: dict, bases: pd.DataFrame):
        """
        Initialise an external instrument model.

        Args:
            dispersion (dict): A dictionary contained the dispersion function sampled on a high resolution grid
                (simple spline interpolation will be used to estimate the dispersion at different locations).
            response (dict): A dictionary containing the response curve sampled on a high resolution grid (simple spline
                interpolation will be used to estimate the dispersion at different locations).
            bases (DataFrame): A DataFrame containing the definition of the inverse bases, required to reconstruct the
                absolute spectrum.
        """
        self.dispersion = dispersion
        self.response = response
        self.bases = bases

    @classmethod
    def from_config_csv(cls, dispersion_path: str, response_path: str, bases_path: str):
        """
        Create an external calibration instrument model from the input configuration files.

        Args:
            dispersion_path (str): Path to the configuration file containing the dispersion.
            response_path (str): Path to the configuration file containing the response.
            bases_path (str): Path to the configuration file containing the inverse bases.

        Returns:
            ExternalInstrumentModel: An external calibration instrument model object.
        """
        # Dispersion
        _disp = np.genfromtxt(dispersion_path, delimiter=',')
        dispersion = dict(zip(['wavelength', 'pseudo-wavelength'], [_disp[0], _disp[1]]))
        # Response
        _resp = np.genfromtxt(response_path, delimiter=',')
        response = dict(zip(['wavelength', 'response'], [_resp[0], _resp[1]]))
        # Bases
        bases, _ = InverseBasesParser().parse_file(bases_path)
        bases = bases.iloc[0]
        bases['inverseBasesCoefficients'] = bases['inverseBasesCoefficients'].reshape(bases['nBases'],
                                                                                      bases[
                                                                                          'nInverseBasesCoefficients'])
        bases['transformationMatrix'] = bases['transformationMatrix'].reshape(bases['nBases'],
                                                                              bases['nTransformedBases'])
        return cls(dispersion, response, bases)

    def get_response(self, wavelength: float) -> np.ndarray:
        """
        Get the response of the mean instrument at a certain wavelength.

        Args:
            wavelength (float): The absolute wavelength.

        Returns:
            ndarray: The response of the mean instrument at the input wavelength.
        """
        tck = interpolate.splrep(self.response.get("wavelength"), self.response.get("response"), s=0)
        return interpolate.splev(wavelength, tck, der=0)

    def wl_to_pwl(self, wavelength: float) -> np.ndarray:
        """
        Convert the input absolute wavelength to a pseudo-wavelength.

        Args:
            wavelength (float): Absolute wavelength.

        Returns:
            ndarray: The corresponding pseudo-wavelength value.
        """
        tck = interpolate.splrep(self.dispersion.get("wavelength"), self.dispersion.get("pseudo-wavelength"), s=0)
        return interpolate.splev(wavelength, tck, der=0)



"""
dispersion_function.py
====================================
Module providing utilities to convert pseudo-wavelength to absolute wavelength and viceversa.
"""

from configparser import ConfigParser
from functools import lru_cache
from os.path import join

import numpy as np
import pandas as pd
from scipy import interpolate

from gaiaxpy.config.paths import config_path, config_ini_file
from gaiaxpy.core.satellite import BANDS, BP_WL, RP_WL


@lru_cache(maxsize=None)
def read_config_file():
    config_parser = ConfigParser()
    config_parser.read(config_ini_file)
    config_file = join(config_path, config_parser.get('core', 'dispersion_function'))
    return pd.read_csv(config_file)


@lru_cache(maxsize=None)
def generate_bp_conversion():
    df = read_config_file()
    wl = df['wl_nm'].copy()
    pwl = df['bp_pwl'].copy()
    flipped_pwl = np.flipud(df['bp_pwl'].copy())
    flipped_wl = np.flipud(df['wl_nm'].copy())
    bp_pwl_to_wl = interpolate.interp1d(flipped_pwl, flipped_wl, kind='linear', bounds_error=False,
                                        fill_value='extrapolate')
    bp_wl_to_pwl = interpolate.interp1d(wl, pwl, kind='linear', bounds_error=False, fill_value='extrapolate')
    return bp_pwl_to_wl, bp_wl_to_pwl


@lru_cache(maxsize=None)
def generate_rp_conversion():
    df = read_config_file()
    df_not_nan = df[df['rp_pwl'].notna()]
    wl = df_not_nan['wl_nm'].copy()
    pwl = df_not_nan['rp_pwl'].copy()
    rp_pwl_to_wl = interpolate.interp1d(pwl, wl, kind='linear', bounds_error=False, fill_value='extrapolate')
    rp_wl_to_pwl = interpolate.interp1d(wl, pwl, kind='linear', bounds_error=False, fill_value='extrapolate')
    return rp_pwl_to_wl, rp_wl_to_pwl


bp_pwl_to_wl, bp_wl_to_pwl = generate_bp_conversion()
bp_pwl_range = [float(bp_wl_to_pwl(BP_WL.low)), float(bp_wl_to_pwl(BP_WL.high))]
bp_wl_range = [BP_WL.low, BP_WL.high]

rp_pwl_to_wl, rp_wl_to_pwl = generate_rp_conversion()
rp_pwl_range = [float(rp_wl_to_pwl(RP_WL.low)), float(rp_wl_to_pwl(RP_WL.high))]
rp_wl_range = [RP_WL.low, RP_WL.high]


def pwl_to_wl(band, pwl):
    """
    Convert the input pseudo-wavelength value(s) into absolute wavelength for the input band (BP or RP).

    Args:
        band (str): BP or RP.
        pwl (ndarray): 1D array containing the pseudo-wavelengths to be converted.

    Returns:
        (ndarray): The absolute wavelengths corresponding to the input pseudo-wavelength values.

    Raises:
        ValueError: If the band string is not equal to BP or RP.
    """
    if band.lower() == BANDS.bp:
        return bp_pwl_to_wl(pwl)
    elif band.lower() == BANDS.rp:
        return rp_pwl_to_wl(pwl)
    else:
        raise ValueError("Unrecognised input band. Only 'BP' or 'RP' values are valid.")


def wl_to_pwl(band, wl):
    """
    Convert the input wavelength value(s) into pseudo-wavelength for the input band (BP or RP).

    Args:
        band (str): BP or RP.
        wl (ndarray): 1D array containing the absolute wavelengths to be converted.

    Returns:
        (ndarray): The pseudo-wavelengths corresponding to the input absolute wavelength values.

    Raises:
        ValueError: If the band string is not equal to BP or RP.
    """
    if band.lower() == BANDS.bp:
        return bp_wl_to_pwl(wl)
    elif band.lower() == BANDS.rp:
        return rp_wl_to_pwl(wl)
    else:
        raise ValueError("Unrecognised input band. Only 'BP' or 'RP' values are recognised.")


def pwl_range(band):
    """
    Return the range where the dispersion function is interpolated. Outside this range extrapolation will occur.

    Args:
        band (str): BP or RP.

    Returns:
        (list): The minimum and maximum values of the pseudo-wavelength range.

    Raises:
        ValueError: If the band string is not equal to BP or RP.
    """
    if band.lower() == BANDS.bp:
        return bp_pwl_range
    elif band.lower() == BANDS.rp:
        return rp_pwl_range
    else:
        raise ValueError("Unrecognised input band. Only 'BP' or 'RP' values are recognised.")


def wl_range(band):
    """
    Return the range where the dispersion function is interpolated. Outside this range extrapolation will occur.

    Args:
        band (str): BP or RP.

    Returns:
        (list): The minimum and maximum values of the wavelength range.

    Raises:
        ValueError: If the band string is not equal to BP or RP.
    """
    if band.lower() == BANDS.bp:
        return bp_wl_range
    elif band.lower() == BANDS.rp:
        return rp_wl_range
    else:
        raise ValueError("Unrecognised input band. Only 'BP' or 'RP' values are recognised.")



"""
xp_continuous_spectrum.py
====================================
Module to represent a BP/RP continuous spectrum.
"""

import numpy as np

from gaiaxpy.core.generic_functions import array_to_symmetric_matrix
from .utils import _list_to_array, get_covariance_matrix
from .xp_spectrum import XpSpectrum


class XpContinuousSpectrum(XpSpectrum):
    """
    A Gaia BP/RP spectrum represented as a continuous function defined as the sum of a set of bases functions multiplied
        by a set of coefficient. This definition is the result of the least squares fit. Covariance and standard
        deviation for the least square solution are also part of the continuous spectrum definition to allow estimating
        errors.
    """

    def __init__(self, source_id, xp, coefficients, covariance, standard_deviation):
        """
        Initialise XP continuous spectrum.

        Args:
            source_id (str): Source identifier.
            xp (str): Gaia photometer, can be either 'bp' or 'rp'.
            coefficients (ndarray): 1D array containing the coefficients multiplying the basis functions.
            covariance (ndarray): 2D array containing the covariance of the least squares solution.
            standard_deviation (float): Standard deviation of the least squares solution.
        """
        XpSpectrum.__init__(self, source_id, xp)
        self.coefficients = coefficients
        self.covariance = covariance
        self.standard_deviation = standard_deviation

    @classmethod
    def from_data_frame(cls, df, band):
        """
        Initialise XP continuous spectrum from a Pandas DataFrame.

        Args:
            df (DataFrame): DataFrame containing at least the fields source_id, BAND_n_parameters, BAND_coefficients,
                BAND_coefficient_correlations, BAND_standard_deviation, where BAND is either 'bp' or 'rp'. The same
                structure as used in the archive for the correlation matrix is expected.
            band (str): Gaia photometer, can be either 'bp' or 'rp'.
        """
        corr = array_to_symmetric_matrix(df[f'{band}_coefficient_correlations'], df[f'{band}_n_parameters'])
        df[f'{band}_coefficient_correlations'] = corr
        cov = get_covariance_matrix(df, band)
        return cls(df['source_id'], band, df[f'{band}_coefficients'], cov, df[f'{band}_standard_deviation'])

    @classmethod
    def get_units(cls):
        return dict()

    def get_coefficients(self):
        """
        Get the coefficients associated with the spectrum.

        Returns:
            ndarray: The 1D array of the coefficients multiplying the basis functions.
        """
        return self.coefficients

    def get_covariance(self):
        """
        Get the covariance associated with the spectrum.

        Returns:
            ndarray: The 2D array of the covariance matrix.
        """
        return self.covariance

    def get_standard_deviation(self):
        """
        Get the standard deviation associated with the spectrum.

        Returns:
            float: The standard deviation of the least squares solution.
        """
        return self.standard_deviation

    def spectrum_to_dict(self):
        """
        Represent spectrum as dictionary.

        Returns:
            dict: A dictionary populated with the minimum set of parameters that need to be stored for this object.
                This is optimised for writing large number of sampled spectra and for this reason the array of positions
                is NOT included as it is expected to be the same for a batch of spectra. The array fo positions can be
                retrieved calling the sampling_to_dict method.
        """
        diagonal = np.sqrt(np.diag(self.covariance))
        diagonal_inv = np.diag(1.0 / diagonal)
        correlation_matrix = np.matmul(np.matmul(diagonal_inv, self.covariance), diagonal_inv)
        return {
            'source_id': self.source_id,
            'xp': self.xp.upper(),
            'standard_deviation': self.standard_deviation,
            'coefficients': _list_to_array(self.coefficients),
            'coefficient_correlations': _list_to_array(_extract_lower_triangle(correlation_matrix)),
            'coefficient_errors': _list_to_array(diagonal),
            'n_parameters': len(self.coefficients)
        }


def _extract_lower_triangle(matrix):
    """
    Extract the lower triangle of the matrix without including the diagonal.
    """
    # Get the indices
    indices = np.tril_indices(matrix.shape[0], k=-1)
    return np.array([matrix[i][j] for i, j in zip(indices[0], indices[1])])



from os.path import isfile
from pathlib import Path

import pandas as pd

from .dataframe_reader import DataFrameReader
from .file_reader import FileParserSelector
from .hdfs_reader import HDFSReader
from .list_reader import ListReader
from .local_file_reader import LocalFileReader
from .query_reader import QueryReader

default_extension = 'csv'


class InputReader(object):

    def __init__(self, content, function, truncation, additional_columns=None, selector=None, disable_info=False,
                 user=None, password=None):
        if additional_columns is None:
            additional_columns = dict()
        self.additional_columns = additional_columns
        if not isinstance(self.additional_columns, dict):
            raise ValueError(f'Additional columns is {type(self.additional_columns)}.')
        self.selector = selector if selector is None else selector
        self.content = content
        self.function = function
        self.truncation = truncation
        self.disable_info = disable_info
        self.user = user
        self.password = password

    def read(self):
        content = self.content
        function = self.function
        truncation = self.truncation
        disable_info = self.disable_info
        additional_columns = self.additional_columns
        selector = self.selector
        # Input data directly provided by the user
        if isinstance(content, pd.DataFrame):
            reader = DataFrameReader(content, function, truncation, additional_columns=additional_columns,
                                     selector=selector, disable_info=disable_info)
        elif (isinstance(content, Path) or isinstance(content, str)) and isfile(content):
            parser = FileParserSelector(function)
            reader = LocalFileReader(parser, content, truncation, additional_columns=additional_columns,
                                     selector=selector, disable_info=disable_info)
        # Actual input data got from the Archive
        elif isinstance(content, list):
            reader = ListReader(content, function, truncation, user=self.user, password=self.password,
                                additional_columns=additional_columns, selector=selector, disable_info=disable_info)
        elif isinstance(content, str) and content.lower().startswith('select'):
            reader = QueryReader(content, function, truncation, user=self.user, password=self.password,
                                 additional_columns=additional_columns, selector=selector, disable_info=disable_info)
        elif isinstance(content, str) and content.lower().startswith('hdfs://'):
            parser = FileParserSelector(function)
            reader = HDFSReader(parser, content, truncation, additional_columns=additional_columns, selector=selector,
                                disable_info=disable_info)
        else:
            raise ValueError('The input provided does not match any of the expected input types.')
        parsed_data, extension = reader.read()
        extension = default_extension if extension is None else extension
        return parsed_data, extension



"""
regular_photometric_system.py
====================================
Module to represent a regular photometric system.
"""

from .internal_photometric_system import InternalPhotometricSystem


class RegularPhotometricSystem(InternalPhotometricSystem):

    def __init__(self, name, config_file=None):
        """
        A photometric system is defined by the set of bands available.
        Args:
            config_file (str): Path to configuration file.
            name (str): Name of the PhotometricSystem
        """
        super().__init__(name, config_file)

    def _correct_flux(self, flux):
        return flux

    def _correct_error(self, flux, error):
        return error



import numpy as np
import pandas as pd

from gaiaxpy.core.generic_functions import array_to_symmetric_matrix, rename_with_required
from .dataframe_numpy_array_reader import DataFrameNumPyArrayReader
from .dataframe_string_array_reader import DataFrameStringArrayReader
from .required_columns import MANDATORY_INPUT_COLS, COV_INPUT_COLUMNS, CORR_INPUT_COLUMNS, TRUNCATION_COLS
from ..core.custom_errors import SelectorNotImplementedError
from ..core.input_validator import check_column_overwrite
from ..core.satellite import BANDS
from ..file_parser.cast import _cast
from ..spectrum.utils import get_covariance_matrix

covariance_columns = ['bp_covariance_matrix', 'rp_covariance_matrix']
matrix_columns = [('bp_n_parameters', 'bp_coefficient_correlations'),
                  ('rp_n_parameters', 'rp_coefficient_correlations')]


def extremes_are_enclosing(row, column):
    return (row[column][0] == '(' and row[column][-1] == ')') or (row[column][0] == '[' and row[column][-1] == ']')


def needs_matrix_conversion(array_columns):
    columns_to_matrix = (column for value, column in matrix_columns)
    return set(columns_to_matrix).intersection(set(array_columns))


class DataFrameReader(object):

    def __init__(self, content, function, truncation, additional_columns=None, selector=None, disable_info=False):
        if not isinstance(content, pd.DataFrame):
            raise ValueError('Input to read must be a DataFrame.')
        if selector is not None:
            raise SelectorNotImplementedError('DataFrame')
        self.additional_columns = dict() if additional_columns is None else additional_columns
        self.content = content.copy()
        self.function_name = function if isinstance(function, str) else function.__name__
        self.disable_info = disable_info
        self.info_msg = 'Reading input DataFrame...'
        self.columns = self.content.columns
        mandatory_columns = MANDATORY_INPUT_COLS.get(self.function_name, list())
        style_columns = list()
        if mandatory_columns:
            style_columns = COV_INPUT_COLUMNS if all([c in mandatory_columns for c in COV_INPUT_COLUMNS]) \
                else CORR_INPUT_COLUMNS
        self.required_columns = mandatory_columns + style_columns
        if truncation:
            self.required_columns = self.required_columns + TRUNCATION_COLS
        self.requested_columns = self.required_columns
        if self.additional_columns:
            check_column_overwrite(additional_columns, self.required_columns)
            self.requested_columns = self.required_columns + [c for c in self.additional_columns.keys() if c not in
                                                              self.required_columns]

    def __get_parseable_columns(self):
        str_columns, np_columns = [], []
        content = self.content
        rows = content.iloc[0:2]
        rows_dict = rows.to_dict('records')
        for row in rows_dict:
            for column in content.columns:
                if isinstance(row[column], str) and extremes_are_enclosing(row, column):
                    str_columns.append(column)
                if isinstance(row[column], np.ndarray):
                    np_columns.append(column)
        return list(set(str_columns)), list(set(np_columns))

    def read(self):
        if not self.disable_info:
            self.show_info_msg()
        content = self.content
        str_array_columns, np_array_columns = self.__get_parseable_columns()
        if str_array_columns:
            data = DataFrameStringArrayReader(content, str_array_columns).read()  # Call string reader
            array_columns = str_array_columns
        elif np_array_columns:
            data = DataFrameNumPyArrayReader(content, np_array_columns).read()
            array_columns = np_array_columns
        else:
            data = content
            array_columns = []
        if needs_matrix_conversion(array_columns):
            for size_column, values_column in matrix_columns:
                data[values_column] = data.apply(lambda row: array_to_symmetric_matrix(row[values_column],
                                                                                       row[size_column]), axis=1)
            if matrix_columns:
                for band in BANDS:
                    data[f'{band}_covariance_matrix'] = data.apply(get_covariance_matrix, axis=1, args=(band,))
                self.requested_columns = self.requested_columns + covariance_columns
        if not self.disable_info:
            self.show_info_msg(done=True)
        data = _cast(data)
        if self.additional_columns:
            data = rename_with_required(data, self.additional_columns)
        data = data[self.requested_columns] if self.requested_columns else data
        # No extension returned for DataFrames
        return data, None

    def show_info_msg(self, done=False):
        msg = self.info_msg
        if done:
            msg = msg + ' Done!'
        print(msg, end='\r')



"""
error_correction.py
====================================
Module that implements the error correction over a multi-photometry.
"""

from functools import lru_cache
from os import listdir
from os.path import join, isfile
from sys import stdout

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from tqdm import tqdm

from gaiaxpy.config.paths import correction_tables_path
from gaiaxpy.core.generic_functions import cast_output, _extract_systems_from_data, _warning
from gaiaxpy.core.generic_variables import pbar_colour, pbar_units, pbar_message
from gaiaxpy.input_reader.input_reader import InputReader
from gaiaxpy.output.photometry_data import PhotometryData


@lru_cache(maxsize=None)
def _get_correctable_systems():
    correction_files = listdir(correction_tables_path)
    systems = [filename.split('-')[2] for filename in correction_files]
    return systems


def _read_system_table(system):
    correction_factors_path = join(correction_tables_path, f'DIDREQ-465-{system}-correction-factors.csv')
    if isfile(correction_factors_path):
        correction_table = pd.read_csv(correction_factors_path, float_precision='round_trip')
        correction_table['bin_centre'] = (correction_table['min_Gmag_bin'] + correction_table['max_Gmag_bin']) / 2
        return correction_table
    raise FileNotFoundError(f'No correction table found for system {system}.')


def _get_correction_array(_mag_G_values, system):
    correction_table = _read_system_table(system)
    min_value, max_value = correction_table['min_Gmag_bin'].iloc[0], correction_table['max_Gmag_bin'].iloc[-1]
    floor_mag_dict = correction_table.set_index('min_Gmag_bin').T.to_dict()
    factor_columns = [col for col in correction_table.columns if 'factor_' in col]
    _mag_G_values = _mag_G_values.to_frame(name='mag_G')
    _mag_G_values['row'] = _mag_G_values['mag_G'].map(lambda x: floor_mag_dict.get(float(np.floor(x))))
    return _mag_G_values.apply(_get_correction_factor, args=(correction_table, factor_columns,
                                                             min_value, max_value,), axis=1)


def _get_correction_factor(mag_row, correction_table, factor_columns, min_value, max_value):
    mag = mag_row['mag_G']
    row = mag_row['row']
    if (mag > max_value) or pd.isna(mag):
        return correction_table[factor_columns].iloc[-1].values
    elif mag < min_value:
        return correction_table[factor_columns].iloc[0].values
    bin_centre = row['bin_centre']
    factors = np.array([row[factor] for factor in factor_columns])
    if mag <= bin_centre:
        return factors
    range_row_max_Gmag_bin = row['max_Gmag_bin']
    if bin_centre < mag < range_row_max_Gmag_bin:
        try:
            next_range_row = correction_table[correction_table['min_Gmag_bin'] == range_row_max_Gmag_bin].iloc[0]
            next_factors = next_range_row[factor_columns]
        except IndexError:
            return factors
        return interp1d(np.array([bin_centre, range_row_max_Gmag_bin]), np.vstack([factors, next_factors]), axis=0)(mag)
    # Raise an exception if none of the conditions match
    raise ValueError('Check the variables being used. The program should never fall in this case.')


def _correct_system(system_df, correction_array):
    # Extract error columns
    error_df = system_df[[column for column in system_df.columns if '_error' in column]]
    error_df_columns = error_df.columns
    if len(error_df_columns) != len(correction_array[0]):
        raise ValueError('DataFrames should have the same number of columns.')
    error_df = np.array(error_df)
    correction_array = np.array([x for x in correction_array])
    product_array = error_df * correction_array
    return pd.DataFrame(product_array, columns=error_df_columns)


# The correction can only be applied for the systems present in the config files
__correctable_systems = _get_correctable_systems()


def apply_error_correction(input_multi_photometry, photometric_system=None, output_path='.',
                           output_file='output_corrected_photometry', output_format=None, save_file=True):
    """
    Apply error correction (see Montegriffo et al., 2022, for more details). Infer photometric systems if not specified.

    Args:
        input_multi_photometry (DataFrame): Photometry DataFrame, can contain photometry for one or more systems.
        photometric_system (obj): Desired photometric system or list of photometric systems.
        output_path (str): Path where to save the output data.
        output_file (str): Name of the output file without extension (e.g. 'my_file').
        output_format (str): Desired output format. If no format is given, the output file format will be the same as
            the input file (e.g. 'csv').
        save_file (bool): Whether to save the output in a file. If false, output_format and output_file_name are
            ignored.

    Returns:
        DataFrame: A DataFrame of all synthetic photometry with corrected errors for the systems for which it is
            possible.
    """
    return _apply_error_correction(input_multi_photometry, photometric_system=photometric_system,
                                   output_path=output_path, output_file=output_file, output_format=output_format,
                                   save_file=save_file)


def _apply_error_correction(input_multi_photometry, photometric_system=None, output_path='.',
                            output_file='output_corrected_photometry', output_format=None, save_file=True,
                            disable_info=False):
    __FUNCTION_KEY = 'correction'
    gaia_system = 'GaiaDr3Vega'
    gaia_G_mag_column = f'{gaia_system}_mag_G'
    input_multi_photometry, extension = InputReader(input_multi_photometry, apply_error_correction, False,
                                                    disable_info=disable_info).read()
    # Validate that it is a multi-photometry, but how? First try below:
    if gaia_G_mag_column not in input_multi_photometry.columns:
        raise ValueError('System Gaia_DR3_Vega, required to apply the error correction is not present in the input'
                         ' photometry.')
    columns = list(input_multi_photometry.columns)
    columns.remove('source_id')
    systems_in_data = _extract_systems_from_data(columns, photometric_system)
    # Only correct the systems that can be corrected
    systems = list(set(systems_in_data) & set(__correctable_systems))
    systems_to_skip = set(systems_in_data) - set(__correctable_systems)
    if systems_to_skip and not disable_info:
        print()
    for system in systems_to_skip:
        _warning(f'System {system} does not have a correction table. The program will not apply error correction over'
                 ' this system.')
    for system in tqdm(systems, desc=pbar_message[__FUNCTION_KEY], total=len(systems), unit=pbar_units[__FUNCTION_KEY],
                       leave=False, colour=pbar_colour, file=stdout):
        system_df = input_multi_photometry[[column for column in input_multi_photometry.columns if
                                            (column.startswith(system) and f'{system}Std' not in column) or
                                            column == gaia_G_mag_column]]
        # Get the correction factors for the mag G column
        correction_array = _get_correction_array(system_df[gaia_G_mag_column], system)
        # Correct error magnitudes
        corrected_system = _correct_system(system_df, correction_array)
        # Apply correction to the original input_multi_photometry
        input_multi_photometry.update(corrected_system)
    output_data = PhotometryData(input_multi_photometry)
    output_data.data = cast_output(output_data)
    output_data.save(save_file, output_path, output_file, output_format, extension)
    return input_multi_photometry



"""
xp_sampled_spectrum.py
====================================
Module to represent a BP/RP sampled spectrum.
"""

from numbers import Number

import numpy as np

from .sampled_spectrum import SampledSpectrum
from .utils import _list_to_array
from .xp_spectrum import XpSpectrum
from ..core.generic_functions import correlation_from_covariance, is_array_empty, is_variable_empty


class XpSampledSpectrum(XpSpectrum, SampledSpectrum):
    """
    A Gaia BP/RP spectrum sampled to a given grid of positions.
    """

    def __init__(self, source_id=0, xp=None, pos=None, flux=None, flux_error=None, cov=None, stdev=None):
        XpSpectrum.__init__(self, source_id, xp)
        self.pos = pos
        self.n_samples = None if is_array_empty(self.pos) else len(self.pos)
        self.flux = None if is_array_empty(flux) else flux
        self.error = None if is_array_empty(flux_error) else flux_error
        self.stdev = None if is_variable_empty(stdev) else stdev
        self.covariance = None if is_array_empty(cov) else cov

    @classmethod
    def from_data_frame(cls, df, sampling, band):
        """
        Initialise a sampled spectrum from a Pandas DataFrame.

        Args:
            df (DataFrame): DataFrame containing at least the fields source_id, flux, flux_error, cov The same
                structure as used in the archive for the correlation matrix is expected.
            sampling (ndarray): Given sampling.
            band (str): Gaia photometer, can be either 'bp' or 'rp'.
        """
        return cls(df['source_id'], band, sampling, df['flux'], df['flux_error'], df['cov'])

    @classmethod
    def from_continuous(cls, continuous_spectrum, sampled_basis_functions, truncation=-1, with_correlation=False):
        """
        Initialise a spectrum.

        Args:
            continuous_spectrum (XpContinuousSpectrum): The continuous representation of this spectrum.
            sampled_basis_functions (SampledBasisFunctions): The set of basis functions sampled onto the grid defining
                the resolution of the final sampled spectrum.
            truncation (int): Number of bases to be used for this spectrum. The set of bases functions used for the
                continuous representation of the spectra has been optimised to ensure that the first bases are the ones
                that contribute most. In many cases, the last bases contribution will be below the noise. Truncation of
                the basis function set to preserve only the significant bases is optional. By default, no truncation
                will be applied, i.e. all bases will be used.
            with_correlation (bool): Whether correlation information should be generated.
        """
        if continuous_spectrum and sampled_basis_functions:
            coefficients = continuous_spectrum.get_coefficients()
            design_matrix = sampled_basis_functions.get_design_matrix()
            pos = sampled_basis_functions.get_sampling_grid()
            if is_array_empty(coefficients):
                return cls(continuous_spectrum.get_source_id(), continuous_spectrum.get_xp(), pos)
            else:
                covariance = continuous_spectrum.get_covariance()
                if isinstance(truncation, Number) and truncation > 0:
                    coefficients = coefficients[:truncation]
                    covariance = covariance[:truncation, :truncation]
                    design_matrix = design_matrix[:truncation][:]
                stdev = continuous_spectrum.get_standard_deviation()
                flux = SampledSpectrum._sample_flux(coefficients, design_matrix)
                flux_error = SampledSpectrum._sample_error(covariance, design_matrix, stdev)
                cov = SampledSpectrum._sample_covariance(covariance, design_matrix) if with_correlation else None
                return cls(continuous_spectrum.get_source_id(), continuous_spectrum.get_xp(), pos, flux, flux_error,
                           cov, stdev)
        else:
            raise ValueError('Either the continuous spectrum or the sampled basis functions argument is empty.')

    @classmethod
    def from_sampled(cls, source_id, xp, pos, flux, flux_error, cov=None):
        """
        Initialise a spectrum.

        Args:
        source_id (long): The source identifier.
        xp (object): The photometer enum (BP/RP).
        pos (ndarray): The array of positions (in pseudo-wavelength or wavelength) of the samples.
        flux (ndarray): The flux value of each sample.
        flux_error (ndarray): The uncertainty on the flux value of each sample.
        cov (ndarray): The covariance matrix.
        """
        return cls(source_id, xp, pos, flux, flux_error, cov)

    def spectrum_to_dict(self, with_correlation):
        """
        Represent spectrum as dictionary.

        Returns:
            dict: A dictionary populated with the minimum set of parameters that need to be stored for this object. This
                is optimised for writing large number of sampled spectra and for this reason the array of positions is
                NOT included as it is expected to be the same for a batch of spectra. The array of positions can be
                retrieved by calling the sampling_to_dict method.
        """
        spectrum_dict = {'source_id': self.source_id, 'xp': self.xp.upper(), 'flux': _list_to_array(self.flux),
                         'flux_error': _list_to_array(self.error)}
        if with_correlation:
            full_correlation = correlation_from_covariance(self.covariance)
            spectrum_dict['correlation'] = full_correlation if \
                full_correlation is None else full_correlation[np.tril_indices(full_correlation.shape[0], k=-1)]
            if self.stdev is not None:
                spectrum_dict['standard_deviation'] = self.stdev
        return spectrum_dict

    def _sampling_to_dict(self):
        """
        Represent sampling as dictionary.

        Returns:
            dict: A dictionary populated with the sampling grid used for this spectrum.
        """
        return {'pos': _list_to_array(self.pos)}

    def _get_fluxes(self):
        return self.flux

    def _get_flux_errors(self):
        return self.error

    def get_positions(self):
        return self.pos

    def _get_covariance(self):
        return self.covariance

    @classmethod
    def get_units(cls):
        return {'flux': 'electron/s', 'flux_error': 'electron/s', 'standard_deviation': 'electron/s', 'pos': 'pwl'}

    @classmethod
    def get_flux_label(cls):
        return r'Flux [$\mathrm{e^-\,s^{-1}}$]'

    @classmethod
    def get_position_label(cls):
        return r'Pseudo\mbox{-}wavelength'
